
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	Gerbil
//
//	Copyright (c) 2001, Bruce Moreland.  All rights reserved.
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	This file is part of the Gerbil chess program project.
//
//	Gerbil is free software; you can redistribute it and/or modify it under
//	the terms of the GNU General Public License as published by the Free
//	Software Foundation; either version 2 of the License, or (at your option)
//	any later version.
//
//	Gerbil is distributed in the hope that it will be useful, but WITHOUT ANY
//	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
//	FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
//	details.
//
//	You should have received a copy of the GNU General Public License along
//	with Gerbil; if not, write to the Free Software Foundation, Inc.,
//	59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

#include "engine.hpp"
#include "gproto.hpp"
#include "hash.hpp"
#include "stats.hpp"
#include "board.hpp"
#include "moves.hpp"
#include "search.hpp"
#include "genmagic.hpp"
#include "moves.hpp"
#include "newbook.hpp"
#include "eval.hpp"
#include "uci.hpp"

#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

#ifdef	_DEBUG
static char const s_aszModule[] = __FILE__;
#endif

// --------------------------------------------------------------------------

//  If this is called while a search is running, it won't change the time
//  control that's being used for this move.

void VSetTimeControl(PCON pcon, int cMoves, TM tmBase, TM tmIncr)
{
    switch (cMoves) {
    case 0:
        pcon->tc.tct = tctINCREMENT;
        break;
    case 1:
        pcon->tc.tct = tctFIXED_TIME;
        break;
    default:
        pcon->tc.tct = tctTOURNEY;
        break;
    }
    pcon->tc.cMoves = cMoves;
    pcon->tc.tmBase = tmBase;
    pcon->tc.tmIncr = tmIncr;
    VPrSendComment(" setting time control with cmodes=%d tmbase=%d tmincr=%d",
        cMoves, tmBase, tmIncr);
}

// --------------------------------------------------------------------------

//  See the caveat regarding "VSetTimeControl".

void VSetFixedDepthTimeControl(PCON pcon, int plyDepth)
{
    pcon->tc.plyDepth = plyDepth;
}

// --------------------------------------------------------------------------

//  This is called at the start of a search that's expected to output a move
//  at some point.
//
//  "tmStart" is *not* set by this routine.

void VSetTime(PCON pcon)
{
    Assert(sizeof(long) == sizeof(TM));		// If this fires, increment
											//  time control is broken.
    TM	tmBase;
    TM	tmExtra;
    switch (pcon->tc.tct) {

    case tctINCREMENT:
        if (pcon->tc.tmIncr) {

            //	An increment time control tries not to go below the increment
            //	plus 30 seconds, or 6 times the increment, whichever is more.
            //
            //	If it finds itself below that, it will the increment minus
            //	1/10 of the time by which it is under the target.
            //
            //	If it is over that, it will use the increment plus 1/20 of
            //	the time by which it is over the target.
            //
            //	So the program will drift slowly down to the target, but if
            //	it gets a ways below it, it will tend to come up quickly.
            //
            //	There are a zillion different increment time controls, and
            //	sometimes I could end up with a stupid base time.  I try hard
            //	to make sure the base time doesn't go below zero, which would
            //	be very bad since my time quantity is an unsigned value.
            //
            //	I will try hard to use >= half of the increment.  I will also
            //	make sure to not use more than half the remaining time (which
            //	is taken care of by the code at "lblSet".
            //
            // base = tmus/30 + incr, sliding time for 30 moves
            tmBase = (pcon->ss.tmUs - pcon->tc.tmIncr) / 25 + pcon->tc.tmIncr;
        } else {
            //
            //	A zero-increment time control will use 1/30 of the remaining
            //	time down to 10 minutes, 1/40 down to one minute, and 1/60
            //	after that.
            //
            tmBase = pcon->ss.tmUs / 25;
        }
        //	This code dummy-checks "tmBase", assigns "tmExtra", then assigns
        //	an end-time based upon "pcon->ss.tmStart".
        //
        //	First I check to see if I'm scheduled to eat more than half of
        //	my remaining time.  I put a ceiling at that amount, right off.
        //
        //	Next, I'll assign some emergency time.  This is 3x the base time,
        //	with a ceiling on base + extra of 1/2 the remaining time.  I will
        //	also allocate no emergency time if I have < 30 seconds on the
        //	clock.
        //
        //	So there will be plenty of "extra" time if there is time left,
        //	otherwise there is little or none.
        //
lblSet:	;
        {
            TM maxTime = pcon->ss.tmUs / 2;
            if (tmBase > maxTime)
                tmBase = maxTime;
            tmExtra = tmBase * 3;
            if (tmBase + tmExtra > maxTime)
                tmExtra = maxTime - tmBase;
            pcon->ss.tmEnd = pcon->ss.tmStart + tmBase;
            pcon->ss.tmExtra = tmExtra;
            VPrSendComment(" setting time tmBase=%d tmExtra=%d, tmUs=%d", tmBase, pcon->ss.tmExtra, pcon->ss.tmUs);
        }
        break;
    case tctFIXED_TIME:
        // this is also the time for the last move in n/m
        tmBase = pcon->tc.tmBase;		// Use *all* the time.
        // keep a margin for the lag
        if( tmBase > 800 )
            tmBase -= 200;
        else if( tmBase > 200 )
            tmBase -= 100;
        else
            tmBase /= 2;
        pcon->ss.tmEnd = pcon->ss.tmStart + tmBase;
        pcon->ss.tmExtra = 0;			// No emergency time.
        VPrSendComment(" setting time tmBase=%d tmExtra=%d, tmUs=%d", tmBase, pcon->ss.tmExtra, pcon->ss.tmUs);
        break;
    case tctTOURNEY:
    {
        int	movLeft;

        if(pcon->isUCI ) {
            movLeft = pcon->tc.cMoves;
        } else {
            int movMade = pcon->gc.ccm / 2 + pcon->gc.movStart - 1;
            while (movMade >= pcon->tc.cMoves)
            movMade -= pcon->tc.cMoves;
            movLeft = pcon->tc.cMoves - movMade;
            Assert(movLeft >= 1);
        }
        //
        //	This expression is kind of nasty.  It is:
        //
        //		Time / (3/4 x Moves + 3)
        //
        //	The point of the "3" is to make sure there's some extra time pad
        //	at the beginning, and the point of the 3/4 is to make sure that
        //	earlier moves take a little longer per move.
        //
        //	This expression has been reduced to ...
        //
        //		(4 * Time) / (3 * Moves + 12)
        //
        //	.. by multiplying the top and the bottom by 4.  This helps out
        //	with the integer math.
        //
        tmBase = (4 * pcon->ss.tmUs) / (3 * (movLeft + 1));
        goto lblSet;
    }
    }
}

// --------------------------------------------------------------------------

//	This is called a whole bunch of times via the engine/interface protocol.
//	Given what is known about the time control, the current time, and the
//	search depth, this may set "ptcon->fTimeout", which will eventually finish
//	the search.

//	This routine has a built-in defense against cases where someone might try
//	to force a move before the first ply has been fully considered.  It will
//	simply ignore such cases.

void VCheckTime(PCON pcon)
{
    TM	tmNow;

    if (pcon->ss.plyDepth == 1)	// Can't time out before a one-ply
        return;					//  search is finished.
    tmNow = TmNow();
    if( pcon->smode == smodeTHINK || pcon->smode == smodeANALYZE ) 
        VPrSendUciNodeCount(false, pcon->ss );
    if (pcon->smode != smodeTHINK)
        return;
    if (tmNow >= pcon->ss.tmEnd) {
        //
        //	Time has expired.  Check for a fail-high or fail low and try to
        //	add some time. 
        //
        if ( pcon->ss.tmExtend ) {
            pcon->ss.tmExtend = false;
            if (pcon->ss.tmExtra < 1000) {
                VPrSendComment("adding all extra time : %d\n", pcon->ss.tmExtra);
                pcon->ss.tmEnd += pcon->ss.tmExtra;
                pcon->ss.tmExtra = 0;
            } else {
                unsigned int moreTime = pcon->ss.tmExtra / 2;
                VPrSendComment("adding some more time : %d\n", moreTime);
                pcon->ss.tmEnd += moreTime;
                pcon->ss.tmExtra -= moreTime;
            }
        }
        if (tmNow >= pcon->ss.tmEnd)
            pcon->fTimeout = true;
    }
}

// --------------------------------------------------------------------------

//	This spits a PV out to the interface.
//
//	WARNING: This function must have the root moves already generated.

void VDumpPv(PCON pcon, int ply, int val, int prsa)
{
	char asz[1024];
	char * sz;
	int	i;
    bool inQS = false;
    // if we get a worse score, set the fail_low flag, this will allocate
    // more time if the normal time expires
#ifdef _DEBUG
//    VPrSendComment(" delta score = %d\n", -(pcon->ss.lastVal - val));
#endif
    if( pcon->ss.lastVal - val >= 20 ) {
		pcon->ss.tmExtend = true;
        if( (pcon->ss.plyDepth > 4) && (pcon->ss.lastVal - val >= 50) ) {
            // add time now
            unsigned int moreTime = (3*pcon->ss.tmExtra) / 4;
            VPrSendComment("adding some time : %d\n", moreTime);
            pcon->ss.tmEnd += moreTime;
            pcon->ss.tmExtra -= moreTime;
        }
    }
    else
        pcon->ss.tmExtend = false;
    if( (pcon->ss.plyDepth > 5) && (pcon->ss.currBestMove.m !=  pcon->argste[0].argcmPv[0].m) ) {
        unsigned int moreTime = (1*pcon->ss.tmExtra) / 9;
        pcon->ss.tmEnd += moreTime;
        pcon->ss.tmExtra -= moreTime;
    }
    pcon->ss.lastVal = val;
    pcon->ss.currBestMove = pcon->argste[0].argcmPv[0];
	pcon->ss.val = val;	// Remember the value of this search so far.
	pcon->ss.prsa = prsa;
	if (!pcon->fPost && !pcon->isUCI)	// The interface doesn't want PV's.
		return;
	sz = asz;
	//
	//	In order to do the SAN thing, I need to actually execute the moves
	//	on the board, so I call "VMakeMove" at the end of this.
	//
	//	I assume the moves in the PV are legal (it can't be otherwise).
	//
	for (i = 0; i < pcon->argste[0].ccmPv; i++) {
		SAN	san;
		PCM	pcmPv;
		PSTE	pste;

		pste = pcon->argste + i;
		pcmPv = &pcon->argste[0].argcmPv[i];
        if( !inQS && (pcmPv->cmk == (cmkHASH | 123456)) && (i > 0) ) {
            sz += sprintf(sz, " [qs] ");
            inQS = true;
        }
		if (i)
			bGenMoves(pcon, pste);
        if( pcon->isUCI )
            CmToSz( pcmPv, sz );
        else {
		    CmToSan(pcon, pste, pcmPv, &san);
		    SanToSz(&san, sz);
        }
		sz += strlen(sz);
		if (i + 1 < pcon->argste[0].ccmPv)
			*sz++ = ' ';
#ifdef _DEBUG
        if( true ) {
        if( true ) {
            unsigned int fh = (pcmPv->cmk >> 16) & 0xffff;
            unsigned int fl = (pcmPv->cmk) & 0xffff;
//            sz += sprintf(sz, "(%d) ", pcmPv->cmk);
//            sz += sprintf(sz, "(%d/%d/%d) ", fh, fl, History.seen_count(pcmPv));
            sz += sprintf(sz, "(%.2f/%.2f %d) ", 100.0f*pcmPv->cmk/1024.0f, 100.0f*History.get_ratio(pcmPv)/1024.0f, History.seen_count(pcmPv));
        } else {
            if( pcmPv->cmk & cmkHASH )
			    *sz++ = 'H';
            if( pcmPv->cmk & cmkKILLER )
			    *sz++ = 'K';
            if( pcmPv->cmk & cmkPV )
			    *sz++ = 'P';
            if( pcmPv->cmk & (cmkGOODCAPT|cmkEQUCAPT) )
			    *sz++ = 'C';
            if( pcmPv->cmk & cmkBADCAPT )
			    *sz++ = 'B';
        }
        *sz++ = ' ';
        }
#endif
		bMove(pcon, pste, pcmPv->m);
	}
	//	Undo all of the make-move's.
	//
	for (; --i >= 0;)
		bUndoMove(pcon, pcon->argste + i, pcon->argste[0].argcmPv[i].m);
	*sz++ = '\0';
	VPrSendAnalysis(ply, pcon->ss.plyMaxDepth, val, TmNow() - pcon->ss.tmStart,
		pcon->ss.nodes, prsa, asz);
}


void VDumpLine(const CON *pcon, const STE *pste, const char *comment)
{
	char asz[1024];
	char * sz;

	sz = asz;
    sz[0] = 0;
	//
	//	In order to do the SAN thing, I need to actually execute the moves
	//	on the board, so I call "VMakeMove" at the end of this.
	//
	//	I assume the moves in the PV are legal (it can't be otherwise).
	//
	for (const STE *p = pcon->argste ; p <= pste; p++) {
        CM move;
        move.m = p->lastMove;
        move.cmk = 0;
        if( move.m  == 0 ) {
            strcat(sz, " - ");
		    sz += strlen(sz);
            *sz++ = ' ';
            if( p == pste ) {
                move.m = (p+1)->lastMove;
                move.cmk = 0;
                CmToSz( &move, sz );
            }
        }
        else
            CmToSz( &move, sz );
		sz += strlen(sz);
        *sz++ = ' ';
//        unsigned int fh = (move.cmk >> 16) & 0xffff;
//        unsigned int fl = (move.cmk) & 0xffff;
//        sz += sprintf(sz, "(%d/%d) ", fh, fl);
        *sz = 0;
	}
    VPrSendComment("V %s %s", comment, asz);
}


// --------------------------------------------------------------------------

//	This is called before a search in order to tell the search which moves
//	to try first when iterating through them.

void VSetPv(PCON pcon)
{
	int	i;
	
	for (i = 0; i < pcon->argste[0].ccmPv; i++)
		lastPV[i+1] = pcon->argste[0].argcmPv[i];
	for (; i < csteMAX - 1; i++) {
		lastPV[i+1].m = 0;
	}
}


// --------------------------------------------------------------------------

//	Once the first STE record is set up properly, this cleans up a few
//	details in the record, and sets up the rest of the array properly.
//	

void VFixSte(PCON pcon)
{
    PSTE	pste = &pcon->argste[0];

    pste->fNull = false;
    for (int i = 1; i < csteMAX; i++)
        pcon->argste[i].side = pste->side ^ (i & 1);
    pste->mlist_start = pcon->move_list;
}

// --------------------------------------------------------------------------

//	FIDE rules state that if a position repeats three times, a draw can be
//	claimed.  But the rules also state that two positions with the pieces in
//	the same place, but with different castling or en-passant possibilities,
//	are not the same position.  In order to avoid claiming draws erroneously,
//	I maintain the notion of a "full" key, which includes XOR'd values for
//	various castling flags and en-passant files.
//
//	It's kind of nasty to compute this, because even though an en-passant
//	capture might appear possible, it might not actually be legal.
//
//	This routine generates legal moves for this ply, so if that screws you up,
//	watch out.
//
//	Ordinarily I don't mess with full hash keys, but sometimes I have to.

HASHK HashkFull(PCON pcon, PSTE pste)
{
    HASHK	hashk = pste->hashkPc;

    Assert(pste->castle < cfMAX);
    hashk ^= s_arghashkCf[pste->castle];
    if (pste->isqEnP != isqNIL) {

        bGenMoves(pcon, pste);
        foreachmove(pste,pcm)
            if (Capt(pcm->m) == B_EP || Capt(pcm->m) == W_EP ) {
                hashk ^= s_arghashkEnP[File(To(pcm->m))];
                break;
            }
    }
    return hashk;
}

// --------------------------------------------------------------------------

//	Execute "szMove" on the board, update game context ("pcon->gc"), then set
//	the other "pcon" fields up so it's ready to search the new position.

//	If it doesn't find the move, it returns FALSE.

//	Advancing a move is mostly moving "pcon->argste[1]" on top of
//	"pcon->argste[0]", after a call to "VMakeMove" has set up the new context.
//	This is a little nasty and dangerous, so it's a potential bug source.

//	If you modify "VMakeMove", take this potential problem into account.

//	Be especially careful if you mess with this routine, since bugs in it
//	won't show up in test suites -- they'll happen in games.

bool FAdvance(PCON pcon, char * szMove)
{
    char    *szCastle = "";
	PSTE    pste = pcon->argste;
//    Print(pcon, pste );
    // handle castle (coming from frc notation)
    if( !stricmp(szMove,"o-o") )
        if(pste->side == W )
            szCastle = "e1g1";
        else
            szCastle = "e8g8";
    if( !stricmp(szMove,"o-o-o") )
        if(pste->side == W )
            szCastle = "e1c1";
        else
            szCastle = "e8c8";
	bGenMoves(pcon, pste);
    foreachmove(pste,pcm) {
        char	asz[16];

        CmToSz(pcm, asz);
        if (!stricmp(asz, szMove) || !stricmp(asz, szCastle)) {
            bMove(pcon, pste, pcm->m);
            pste[0] = pste[1];
            VFixSte(pcon);
            // add the move to the list of played move
            pcon->gc.argcm[pcon->gc.ccm] = *pcm;
            ++pcon->gc.ccm;
            pcon->gc.arghashk[pcon->gc.ccm] = pste->hashkPc;
            pcon->gc.arghashkFull[pcon->gc.ccm] = HashkFull(pcon, pste);
            return true;
        }
    }
    VPrSendComment("we have a problem FAdvance failed for %s\n", szMove);
#if 0
    {
        foreachmove(pste,pcm) {
		    char	asz[16];
		    CmToSz(pcm, asz);
            VPrSendComment(" a move =%s\n", asz);
	    }
    }
#endif
	return false;
}

// --------------------------------------------------------------------------

//	This backs up one move in the current game, and deletes all trace of the
//	backed up move.

//	It works by going back to the beginning and going forward.

//	This routine is a little evil because it assumes that parts of the current
//	game context aren't blown out by "FInitCon".

void VUndoMove(PCON pcon)
{
    int	i;
    GAME_CONTEXT	gc;

    gc = pcon->gc;
    if (!gc.ccm)
        return;
    bool f = SetBoardFromFEN(pcon, gc.aszFen);
    Assert(f);
    for (i = 0, gc.ccm--; i < gc.ccm; i++) {
        char	asz[16];

        CmToSz(&gc.argcm[i], asz);
        f = FAdvance(pcon, asz);
        Assert(f);
    }
}

// --------------------------------------------------------------------------

//	This checks to see if the side to move at the root of "pcon" is
//	stalemated.

bool FStalemated(PCON pcon)
{
    PSTE    pste = pcon->argste;
    if (pste->checkf == 0 )
        return false;
    bGenMoves(pcon, pste);
    return legalMoveCount(pste) == 0;
}


// --------------------------------------------------------------------------

//	A simple routine that checks for draws based upon the game history, and by
//	that I mean repetition draws and 50-move draws.  It does nothing about
//	stalemate, insuffient material, etc.
//
//	This uses "full" hash keys to check for 3x repetition, in order to avoid
//	declaring a draw when a position has appeard three times, but with
//	different castling and/or en-passant possibilities.

void VDrawCheck(PCON pcon)
{
	PSTE pste = &pcon->argste[0];
	int	cReps;

	Assert(pcon->smode == smodeTHINK);
	if ((pste->valPnUs == 0) && (pste->valPnThem == 0)) {
		//
		//	If there are no pawns, claim a draw due to insufficient material
		//	if there are no pieces, one side has a bare king and the other
		//	has a minor, or if both sides have bishops (same color).
		//
		if ((pste->valPcUs == 0) && (pste->valPcThem == 0)) {
lblIns:		VPrSendResult("1/2-1/2", "Insufficient material");
			return;
		}
		if ((pste->valPcUs == valBISHOP || pste->valPcUs == valKNIGHT) && (pste->valPcThem == 0))
			goto lblIns;
		if ((pste->valPcUs == 0) && (pste->valPcThem == valBISHOP || pste->valPcThem == valKNIGHT))
			goto lblIns;
		if ((pste->valPcUs == valBISHOP) && (pste->valPcThem == valBISHOP)) {
#if 0
			int	argco[coMAX];
            int co;
			for (co = coBLACK; co <= coWHITE; co++) {
				int	ipi;
				int	rnk;
				int	fil;
				
				for (ipi = 0; ipi < pcon->argcpi[co]; ipi++)
					if ((!pcon->argpi[co][ipi].fDead) &&
						(pcon->argpi[co][ipi].pc == pcBISHOP))
						break;
				if (ipi == pcon->argcpi[co])
					break;
				rnk = RnkFromIsq(pcon->argpi[co][ipi].isq);
				fil = FilFromIsq(pcon->argpi[co][ipi].isq);
				argco[co] = !((rnk ^ fil) & 1);
			}
			if ((co > coWHITE) && (argco[coWHITE] == argco[coBLACK]))
				goto lblIns;
#endif
		}
	}
	if (pcon->argste[0].plyFifty >= 50 * 2) {
		VPrSendResult("1/2-1/2", "Draw by 50-move rule");
		return;
	}
	cReps = 1;
	for (int i = 0; i < pcon->gc.ccm; i++)
		if (pcon->gc.arghashkFull[i] ==
			pcon->gc.arghashkFull[pcon->gc.ccm])
			if (++cReps >= 3) {
				VPrSendResult("1/2-1/2", "Draw by repetition");
				return;
			}
}

// --------------------------------------------------------------------------

static char const * c_argszUsage[] = {
	"Usage: %s [flags]",
	"  -?        | Usage",
    "  -b<B>     | {0 | 1} disable or enable the use of an opening book",
	"  -bf<F>    | name of the opening book to use",
	"  -hp<N>    | size of the pawn hash table in bytes",
	"  -ht<N>    | size of the transposition tables in bytes",
	"  -p        | Tell the engine to reduce its system priority.",
    "  -r<B>     | {0 | 1} disable of enable the auto resign feature",
	"  -t<F> <N> | Profile a position for N seconds, F is the FEN string",
	NULL,
};

static void VUsage(int argc, char * argv[])
{
	for (int i = 0; c_argszUsage[i] != NULL; i++)
		VPrSendComment(c_argszUsage[i], argv[0]);
}

// --------------------------------------------------------------------------

PCON PconInitEngine(int argc, char * argv[])
{
	static	CON s_con;
	PCON pcon = &s_con;
	char	aszBook[512];	// Opening book currently loaded.
	char	newBook[512];	// Opening book currently loaded.
	char	egbbdir[512];	// end game bitbases directory
	int	cbMaxHash;			// Number of transposition hash element bytes
							//  at most (actual number will be <= this).
	int	cbMaxPawnHash;		// Number of pawn hash bytes at most (actual
							//  number will be <= this).
	int	bdm;				// Book dump mode.
	bool	f;
	char	aszProfileFen[256];
	TM	tmProfile = 1;

    cyranoOptions.init();

	bdm = bdmNONE;
	VGetProfileSz("OldBookFile", "gerbil.opn", aszBook, sizeof(aszBook));
	VGetProfileSz("BookFile", "test.pbk", newBook, sizeof(newBook));
	VGetProfileSz("egbbdir", ".", egbbdir, sizeof(egbbdir));
    bool fUseBook = IGetProfileInt("UseBook", 1) ? true : false;
	cbMaxHash = IGetProfileInt("MaxHash", 32);
	cbMaxPawnHash = IGetProfileInt("MaxPawnHash", 1000000);

    pcon->fResign = IGetProfileInt("Resign", 1) ? true : false;
	pcon->fLowPriority = false;
	aszProfileFen[0] = '\0';
	//
	//	Process command line switches after setting appropriate defaults.
	//
	for (int i = 1; i < argc; i++) {
		char * sz = argv[i];

		switch (*sz++) {
		case '/':
		case '-':
			switch (*sz++) {
			case '?':
lblError:		VUsage(argc, argv);
				return NULL;
			case 'b':
				switch (*sz++) {
				case 'f':		// -bf[F] "BookFile"
					if (*sz != '\0')
						strcpy(newBook, sz);
					else if (i + 1 == argc)
						goto lblError;
					else if ((argv[++i][0] != '-') &&
						(argv[i][0] != '/'))
						strcpy(newBook, argv[i]);
					else
						goto lblError;
					break;
				case 's':		// -bs[F] "Bitbase"
					if (*sz != '\0')
						strcpy(egbbdir, sz);
					else if (i + 1 == argc)
						goto lblError;
					else if ((argv[++i][0] != '-') &&
						(argv[i][0] != '/'))
						strcpy(egbbdir, argv[i]);
					else
						goto lblError;
					break;
				default:		// -b[D] "UseBook"
					if (*--sz != '\0') {
						if (!isdigit(*sz))
							goto lblError;
                        fUseBook = atoi(sz) ? true : false;
					} else if (i + 1 == argc)
						goto lblError;
					else if (isdigit(argv[++i][0]))
                        fUseBook = atoi(argv[i]) ? true : false;
					else
						goto lblError;
					break;
				}
				break;
			case 'd':
				switch (*sz++) {
				case 'f':
					bdm = bdmFULL;
					break;
				case 'c':
					bdm = bdmCOMPACT;
					break;
				default:
					goto lblError;
				}
				break;
			case 'h':
				switch (*sz++) {
				case 'p':		// -hp[D] "MaxPawnHash"
					if (*sz != '\0') {
						if (!isdigit(*sz))
							goto lblError;
						cbMaxPawnHash = atoi(sz);
					} else if (i + 1 == argc)
						goto lblError;
					else if (isdigit(argv[++i][0]))
						cbMaxPawnHash = atoi(argv[i]);
					else
						goto lblError;
					break;
				case 't':		// -hp[D] "MaxHash"
					if (*sz != '\0') {
						if (!isdigit(*sz))
							goto lblError;
						cbMaxHash = atoi(sz);
					} else if (i + 1 == argc)
						goto lblError;
					else if (isdigit(argv[++i][0]))
						cbMaxHash = atoi(argv[i]);
					else
						goto lblError;
					break;
				default:
					goto lblError;
				}
				break;
			case 'p':			// -p "Priority"
				pcon->fLowPriority = true;
				break;
			case 'r':			// -r<B> "Resign"
				if (*sz != '\0') {
					if (!isdigit(*sz))
						goto lblError;
                    pcon->fResign = atoi(sz) ? true : false;
				} else if (i + 1 == argc)
					goto lblError;
				else if (isdigit(argv[++i][0]))
                    pcon->fResign = atoi(argv[i]) ? true : false;
				else
					goto lblError;
				break;
			case 't':			// -t<F> <N> "Profile Fen"
				if (*sz != '\0')
					strcpy(aszProfileFen, sz);
				else if (i + 1 == argc)
					goto lblError;
				else if ((argv[++i][0] != '-') && (argv[i][0] != '/'))
					strcpy(aszProfileFen, argv[i]);
				else
					goto lblError;
				if (i + 1 == argc)
					goto lblError;
				else if (!isdigit(argv[++i][0]))
					goto lblError;
				tmProfile = atol(argv[i]) * 1000;
				break;
			default:
				goto lblError;
			}
			break;
		default:
			goto lblError;
		}
	}
	if (bdm == bdmNONE)	// Avoid writing output if I'm going to dump the book.
		VPrSendComment("UseBook: %s", c_argszNoYes[fUseBook]);
    InitBitboards();
	VInitHashk();
	VInitMiddlegamePawnTable();
    uciHash.set_spin( cbMaxHash );
	if (!FInitHashp(pcon, cbMaxPawnHash))
		return NULL;
    book_clear();
    uciOwnBook.set_check( fUseBook );
    uciBookName.set_string( newBook );
    uciEGBBDir.set_string( egbbdir );
    if( fUseBook )
        book_open( newBook );
	VReseed();
	pcon->smode = smodeIDLE;
	f = SetBoardFromFEN(pcon, s_aszFenDefault);
	Assert(f);
    {
    	PSTE	pste = pcon->argste;
        initEG(pcon, pste);
    	f = SetBoardFromFEN(pcon, s_aszFenDefault);
    }
//    book_find_move(pcon, &pcon->argste[0], false);
//    book_disp(pcon, &pcon->argste[0]);
	VPrSetPost(pcon, false);
	VPrSetPonder(pcon, false);
    VPrSetUciMode(pcon, false);
	VSetTimeControl(pcon, 1, 2000, 0);	// Set to play 2 seconds per move if
										//  the user is foolish enough to try
										//  to run the program from the
										//  command line.
//	VSetTimeControl(pcon, 0, 9999, 9999);	// Set to play 2 seconds per move if
	if (aszProfileFen[0]) {
		if (!FPrSetboard(pcon, aszProfileFen)) {
			VPrSendComment("Can't initialize profile FEN.");
			exit(1);
		}
		VPrSetTimeControl(pcon, 1, tmProfile, 0);
		VPrSetPost(pcon, false);
		VPrSetPonder(pcon, false);
		VPrSendComment("Profile for %ld sec", tmProfile / 1000);
		VPrThink(pcon, tmProfile, 0);
		exit(1);
	}
	return pcon;
}

// --------------------------------------------------------------------------


void VStalemate(int coStalemated)
{
    VPrSendResult("1/2-1/2", "Stalemate");
}

//	"coWin" just checkmated.  Tell the interface.

void VMates(int coWin)
{
    char	aszResult[64];
    char	aszReason[64];

    sprintf(aszResult, "%d-%d", coWin, coWin ^ 1);
    sprintf(aszReason, "%s mates", s_argszCo[coWin]);
    VPrSendResult(aszResult, aszReason);
}

// --------------------------------------------------------------------------

//	These two functions look for specific moves in the move list, in order to
//	increase their sort key so they will sort first.

void SortPV(PSTE pste, int depth)
{
    foreachmove(pste, pcm)
		if ( *pcm == lastPV[depth] ) {
			pcm->cmk |= cmkPV;
			return;
		}
}

void SortHash(PSTE pste, PHASH phash)
{
    U32 hm = phash ? phash->getMove() : 0;
    if( !hm )
        return;
    foreachmove(pste, pcm)
        if (pcm->m == hm) {
			pcm->cmk |= cmkHASH;
			return;
		}
}


// --------------------------------------------------------------------------
