//
// Cyrano Chess engine
//
// Copyright (C) 2007  Harald JOHNSEN
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
//
//


#include "engine.hpp"
#include "moves.hpp"
#include "magicmoves.h"
#include "genmagic.hpp"
#include "eval.hpp"
#include "stats.hpp"
#include "hash.hpp"
#include <stdlib.h>     // for dummy rand()

#ifdef _DEBUG
static char const s_aszModule[] = __FILE__;
#endif


const int promotionSqrFileA[2] = {A1,A8};
const int promotionSqrFileH[2] = {H1,H8};

static bitboard attackMap[coMAX];
static int posval[coMAX];       // position value
static int tacval[coMAX];       // tactical value
static int otherval[coMAX];     // other value ?
static int endgval[coMAX];      // end game values
static int notpawn[coMAX];
static int attackval[coMAX];
static int c_mobility[coMAX];
static int mobility[coMAX];
static int defSpace[coMAX];
static int pawnDist[coMAX];
static bitboard knight_moves[coMAX];
static bitboard tmp_attack[coMAX];
static int bishopSquare[coMAX];
static PHASHPawn phashp;
static int mat[coMAX];
static int underdev[coMAX];
//
static bitboard kingRing[coMAX];
static bool kingAtkFSlider[coMAX];
static int c_kingAtk[coMAX];
static int c_kingDef[coMAX];
static int safetyval[coMAX];
//
static bitboard moveBlocker[coMAX];
static bitboard targets[coMAX];
static int c_attack[coMAX];
//
static bitboard piece_attacks[coMAX][16];
static bitboard piece_list[coMAX][16];
static int piece_list_sqr[coMAX][16];
static U8 piece_list_type[coMAX][16];
static int c_piece_attacks[coMAX];
static bitboard attacked_material[coMAX];
static bitboard busy_pieces;
static int c_piece_mob[coMAX];

static struct mobDebug {
    int piece;
    int mobility;
} mobDebugTable[32];
static int mobDebugIdx;

//==== ATTACK TABLE ============================================================

// I only look at the attacker type
//U8 attackMap[2][64];
typedef struct _posdir {
    U8 pos;
    U8 dir;
} posdir;
//static posdir PosDirTab[8][64][64];

// attacker type sorted by value
#define ATK_PAWN    2
#define ATK_MINOR   2
#define ATK_ROOK    2
#define ATK_QUEEN   3

//#define ATK_KING    0x01

/* piece code defines */
#define APC_no_piece 0
#define APC_pawn 1
#define APC_knight 2
#define APC_bishop 3
#define APC_rook 4
#define APC_queen 5
#define APC_king 6
#define APC_bpawn 7
#define APC_BLACKKING   12

//const int KING		= 0;
//const int QUEEN		= 2;
//const int ROOK		= 4;
//const int BISHOP	= 6;
//const int KNIGHT	= 8;
//const int PAWN		= 10;

// mobility counter increment depending on square content
// t[12] = black king
//                      -,WK,BQ,WQ,BR,WR,BB,WB,BN,WN,BP,WP,BK
// 0 is empty, BK replaced with 12 in pos[]
//const U8 mobCount[2][16] = {{1,1,0,1,0,1,0,1,0,1,0,1,0},
//                            {1,0,1,0,1,0,1,0,1,0,1,0,1}};

static char stringPc[] = "KkQqRrBbNnPp";

//#define mobK(n)     (n)
#define mobK(n)     ((3*n)/4)   // reduce since no safety eval
#define queenK(n)   ((3*mobK(n))/4)
#define rookK(n)   ((3*mobK(n))/4)
#if 1

// mobility adjusted with unsafe mobility
static int KnightMobility[1+8] = {mobK(-15),mobK(-10),mobK(-5),mobK(0),mobK(7),
    mobK(12),mobK(15),mobK(15),mobK(15)};

// 14 * ((64-32) / 64) => 7,        14 * ((64-20)/64) => 9.6
static int BishopMobility[1+14] = {mobK(-15),mobK(-12),mobK(-10),mobK(-7),mobK(-4),mobK(0),mobK(5),
    mobK(10),mobK(12),mobK(15),mobK(20),mobK(20),mobK(20),mobK(20),mobK(20)};

// first and last ranks don't really count
// 14 * ((48-32)/48)   => 4.6,      14 * ((48-20)/48) => 8.1
//static int RookMobility[1+14] = {mobK(-30),mobK(-20),mobK(-10),mobK(-5),mobK(0),mobK(5),mobK(10),
//    mobK(15),mobK(20),mobK(22),mobK(25),mobK(27),mobK(30),mobK(30),mobK(30)};
static int RookMobility[1+14] = {rookK(-20),rookK(-10),rookK(-5),rookK(0),rookK(5),rookK(10),
    rookK(15),rookK(20),rookK(22),rookK(25),rookK(25),rookK(25),rookK(25),rookK(25),rookK(25)};

// bishop+rook 32 => 11.6       20 => 17.7
static int QueenMobility[1+28] = {queenK(-25),queenK(-22),queenK(-17),queenK(-12),queenK(-8),queenK(-4),queenK(0),queenK(4),
/*8*/    queenK(8),queenK(12),queenK(15),queenK(16),queenK(17),queenK(18),queenK(19),queenK(20),queenK(21),queenK(22),queenK(23),
/*19*/    queenK(23),queenK(24),queenK(25),queenK(26),queenK(27),queenK(28),queenK(30),queenK(30),queenK(30),queenK(30)};

#else
// base value
static int KnightMobility[1+8] = {mobK(-30),mobK(-20),mobK(-10),mobK(0),mobK(10),
    mobK(15),mobK(20),mobK(22),mobK(25)};

// 14 * ((64-32) / 64) => 7,        14 * ((64-20)/64) => 9.6
static int BishopMobility[1+14] = {mobK(-30),mobK(-25),mobK(-20),mobK(-15),mobK(-10),mobK(-5),mobK(0),
    mobK(5),mobK(10),mobK(15),mobK(20),mobK(25),mobK(30),mobK(32),mobK(35)};

// first and last ranks don't really count
// 14 * ((48-32)/48)   => 4.6,      14 * ((48-20)/48) => 8.1
static int RookMobility[1+14] = {mobK(-30),mobK(-20),mobK(-15),mobK(-10),mobK(-5),mobK(0),mobK(10),
    mobK(15),mobK(20),mobK(22),mobK(25),mobK(27),mobK(30),mobK(30),mobK(30)};

// bishop+rook 32 => 11.6       20 => 17.7
static int QueenMobility[1+28] = {queenK(-50),queenK(-40),queenK(-30),queenK(-20),queenK(-10),queenK(-7),queenK(-6),queenK(-5),
    queenK(-4),queenK(-3),queenK(-2),queenK(0),queenK(2),queenK(5),queenK(7),queenK(10),queenK(12),queenK(15),queenK(17),
    queenK(20),queenK(22),queenK(25),queenK(27),queenK(30),queenK(30),queenK(30),queenK(30),queenK(30),queenK(30)};
#endif

//==============================================================================

static const bitboard rankMask[8] = {0xff, 0xff00, 0xff0000, 0xff000000, 0xff00000000, 0xff0000000000, 0xff000000000000, 0xff00000000000000};
const unsigned char castle_mask[2] = { 12, 3 };

// I don't want to use one array per color for positional values
// this array precomputes rank swap
int const remapSq[2][64] = {
    {
//  a   b   c   d   e   f   g   h
	0,	1,	2,	3,	4,	5,	6,	7,
	8,	9,	10,	11,	12,	13,	14,	15,
	16,	17,	18,	19,	20,	21,	22,	23,
	24,	25,	26,	27,	28,	29,	30,	31,
	32,	33,	34,	35,	36,	37,	38,	39,
	40,	41,	42,	43,	44,	45,	46,	47,
	48,	49,	50,	51,	52,	53,	54,	55,
	56,	57,	58,	59,	60,	61,	62,	63
    },{
	56,	57,	58,	59,	60,	61,	62,	63,
	48,	49,	50,	51,	52,	53,	54,	55,
	40,	41,	42,	43,	44,	45,	46,	47,
	32,	33,	34,	35,	36,	37,	38,	39,
	24,	25,	26,	27,	28,	29,	30,	31,
	16,	17,	18,	19,	20,	21,	22,	23,
	8,	9,	10,	11,	12,	13,	14,	15,
	0,	1,	2,	3,	4,	5,	6,	7
    }
};

//
//	These are piece-square tables.  The award bonuses (or penalties) in
//	centipawns if a piece is sitting on a particular square
//

/*
const int KING		= 0;    => 0
const int QUEEN		= 2;    => 1
const int ROOK		= 4;    => 2
const int BISHOP	= 6;    => 3
const int KNIGHT	= 8;    => 4
const int PAWN		= 10;   => 5
*/
static int const staticPosVal[6+2][64] = {
	//
	//	White king.
	//
//  a   b   c   d   e   f   g   h
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-7,	-15,-15,-15,-15,-15,-15,-7,
	-5,	-5,	-12,-12,-12,-12,-5,	-5,
	 3,	 7,	  9,-12, -9,-12, 12, 10,
	//
	//	White queen.
	//
//  a   b   c   d   e   f   g   h
	-2,	0,	2,	2,	2,	2,	0,	0,
	-2,	8,	8,	8,	8,	8,	8,	0,
	-2,	4,	4,	5,	5,	4,	4,	0,
	0,	0,	0,	2,	2,	0,	0,	0,
	0,	0,	0,	2,	2,	0,	0,	0,
	0,	1,	1,	1,	1,	0,	0,	0,
	0,	0,	1,	1,	1,	0,	0,	0,
	-2,	-2,	-2,	0,	0,	-2,	-2,	-2,
	//
	//	White rook.
	//
//  a   b   c   d   e   f   g   h
	0,	3,	3,	3,	3,	3,	3,	0,
	8,	8,	8,	8,	8,	8,	8,	8,
	-2,	5,	5,	5,	5,	5,	5,	-2,
	-2,	0,	0,	0,	0,	0,	0,	-2,
	-2,	0,	0,	0,	0,	0,	0,	-2,
	-2,	0,	0,	0,	0,	0,	0,	-2,
	-4,	0,	0,	0,	0,	0,	0,	-4,
	-4,	0,	3,	6,	6,	3,	0,	-4,
	//
	//	White bishop.
	//
//  a   b   c   d   e   f   g   h
	5,	5,	5,	5,	5,	5,	5,	5,
	8,	8,	8,  8,  8,	8,	8,	8,
	8,	8,	8,  8,  8,	8,	8,	8,
	8,	9,	8,  8,  8,	8,  9,	8,
	8,	8,  9,  8,  8,  9,	8,	8,
	6,	8,	8,  8,  8,	8,	8,	6,
	4,  7,	5,  3,  7,	5,  7,	4,
	0, -4,	0,	0,	0,	0, -4,	0,
	//
	//	White knight.
	//
//  a   b   c   d   e   f   g   h
	-9,-5,-5, -5, -5, -5, -5, -9,
	-7,	0,  6,  6,  6,  6,	0, -7,
	-4,	5,  7,  7,  7,  7,	5, -4,
	-4,	7,	7,  8,  8,	7,	7, -4,
	-4,	8,	5,	6,	6,	5,	8, -4,
	-4,	5,  6,	5,	5,  6,	5, -4,
	-7,	0,  5,  7,  7,  5,	0, -7,
	-9,-2, -2, -2, -2, -2, -2, -9,
	//
	//	White pawn.
	//
// TODO:not up to date =============================================================
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,  // a8..h8
	0,	0,	0,	0,	0,	0,	0,	0,
	1,	2,	4,  9,  9,	0,	0,	0,
	1,	2,	4,	7,	7,	2,	0,	0,
	2,	2,	5,  7,  7,	4,	0,	0,
	2,	2,	1,	5,	5,	0,	2,	2,
	5,	5,	2,	0, -2,	5,	5,	5,
	0,	0,	0,	0,	0,	0,	0,	0,  // a1..h1
};

//
//


static const int knightOutpostVal[64] = {
	//
	//	White knight.
	//
//  a   b   c   d   e   f   g   h
	 0, 0, 0,  0,  0,  0,  0,  0,
	 0,	0,  5,  7,  7,  5,	0, 0,
	 0,	0,  7,  7,  7,  7,	0, 0,
	 0,	0,	5,  7,  7,	5,	0, 0,
	 0,	0,	3,	5,	5,	3,	0, 0,
	 0, 0, 0,  0,  0,  0,  0,  0,
	 0, 0, 0,  0,  0,  0,  0,  0,
	 0, 0, 0,  0,  0,  0,  0,  0,
};

//	The king is a weird piece because in the engame its eval is approximately
//	opposite of its eval in the middlegame or opening.  At the start of the
//	game, you want to hide your king in the corner, and at the end, you want
//	to run to the middle.

//	This problem is taken care of by having two tables.  If you try to make
//	things work with the middlegame table, the program will play like crap in
//	endings.

static int	const KingEGPosVal[64] = {
	//
	//	White king.
	//
//  a   b   c   d   e   f   g   h
	-5,	0,	0,	0,	0,	0,	0, -5,
	3,	6,	7,	8,	8,	7,	6,  3,
	4,	7,	8,	9,	9,	8,	7, 4,
	3,	6,	7,	8,	8,	7,	6, 3,
	2,	4,	5,	6,	6,	5,	4, 2,
	0,	2,	3,	4,	4,	3,	2, 0,
	-3,	0,	1,	2,	2,	1,	0, -3,
	-5, -2,-1,	-1,	-1,	-1, -2, -5,
};

#define	sideK	0
#define	sideC	1
#define	sideQ	2

#define	sideMAX	3

//	I didn't use the #defines because that is hard to format.  This is an
//	array of char instead of int, which is a decision based upon space.
//
//	Note that c1 is considered a q-side square but f1 is not considered a
//	k-side square.  I want to do everything possible to:
//
//	1.	Convince the program to castle, and O-O-O, which puts the king on
//		c1, is a fine thing to do.
//
//	2.	Prevent the program from playing Kf1.  f1 is not a good square for
//		the king, and I'd rather it play O-O if it's going to move its king
//		into the k-side corner.

static char const c_argside[64] = {
//  a   b   c   d   e   f   g   h
	2,	2,	2,	1,	1,	1,	0,	0,
	2,	2,	1,	1,	1,	1,	0,	0,
	1,	1,	1,	1,	1,	1,	1,	1,
	1,	1,	1,	1,	1,	1,	1,	1,
	1,	1,	1,	1,	1,	1,	1,	1,
	1,	1,	1,	1,	1,	1,	1,	1,
	2,	2,	1,	1,	1,	1,	0,	0,
	2,	2,	2,	1,	1,	1,	0,	0,
};

// contain the distance of a square to the border
// used to eval simple mat positions to speedup the finding of the mat
// higher value when opposite king is near the border

const static int c_argdistborder[64] = {
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	10,	10,	10,	10,	10,	10,	0,
	0,	10,	20,	20,	20,	20,	10,	0,
	0,	10,	20,	30,	30,	20,	10,	0,
	0,	10,	20,	30,	30,	20,	10,	0,
	0,	10,	20,	20,	20,	20,	10,	0,
	0,	10,	10,	10,	10,	10,	10,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
};

// minor value less when there is fewer pawns
static int minor_eg_malus[] = {0,0,75,15,0,0,0,0,0,0};

// rank bonus for advancing pawns
static int pawn_bonus_to_7[coMAX][8] = {{0,20,15,10,5,5,0,0}};
// file bonus for pawns
// this could be included in pst but it's simpler here
static int pawnFileValMG[8] = {-15,0,0,0,0,0,0,-15};
static int pawnFileValEG[8] = {10,7,3,0,0,3,7,10};

// piece:Q+R+R+B+B+N+N + K
// full score when 8 pieces (start of game)
// 0 or near 0 when only the king left
// mid game position tables are worthless in end game
//#define MidGameValue(n,pop)     (((n)*(pop))/8)
//#define EndGameValue(n,pop)     (((n)*(8-pop))/8)
// ------------------------------? K+1+2+3+4+5+6+7 --- space for more pieces...
static int MidGameValueCoef[] = {0,0,0,2,4,5,7,8,8,8,8,8,8,8,8,8,8,8,8};
static int EndGameValueCoef[] = {8,8,8,6,4,2,0,0,0,0,0,0,0,0,0,0,0,0,0};
#define MidGameValue(n,pop)     (((n)*MidGameValueCoef[pop])/8)
#define EndGameValue(n,pop)     (((n)*EndGameValueCoef[pop])/8)

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-


void VInitMiddlegamePawnTable(void)
{
    // TODO:check that ==========================================================
    for(int i = 0; i < 8 ; i++)
        pawn_bonus_to_7[coWHITE][7-i] = pawn_bonus_to_7[coBLACK][i];
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

// from Gerd
static unsigned int colorOfSquare (unsigned int sq) {
   return  (0xAA55AA55 >> sq) & 1;
}
// from h.g.m
static bool sameSquareColor(unsigned int sq1, unsigned int sq2) {
    sq1 ^= sq2;
    return  ( (0xAA55AA55 >> sq1) & 1);
}

static int sqDistance( int isqFrom, int isqTo ) {

    int dr = Rank( isqFrom ) - Rank( isqTo );
    int df = File( isqFrom ) - File( isqTo );
    if( dr < 0 ) dr = -dr;
    if( df < 0 ) df = -df;
    if( dr > df )
        return dr;
    return df;
}
// does not check if there is two bishops
static bool bishopcanreach(const CON *pcon, int side, int sqr ) {
    bitboard bishops = pcon->pieces[BISHOP|side];
    if( bishops == 0 )
        return false;
    int bishop_sqr = LSB( bishops );
    return sameSquareColor(bishop_sqr, sqr);
}

//
// Square rule
// used to see a winning position with KP vs K
// and to give a bonus for passed pawn based on distance to king
//
static bool SquareRuleForPawn(const CON *pcon, const STE *pste, int co, int isqPawn, int & val) {
    // compute ranks from pawn to promotion
    // compute file dist from opposite king to pawn
    // if dist to promote < dist from king then win
    int myturn = pste->side != co;
    int isqKing = pcon->king[co ^ 1];
    int isqWKing = pcon->king[co];
    int distPawn;
    int rankKing = 7 - Rank( isqKing );
    int rankPawn = 7 - Rank( isqPawn );
    val = 0;
    // can not count if own king on pawn file
    if( File(isqWKing) == File(isqPawn) )
        return false;
    if( co != coWHITE ) {
        distPawn = 7 - rankPawn;
        rankKing = 7 - rankKing;
    } else
        distPawn = rankPawn;
    if( rankKing > (distPawn + myturn) ) {
        val = 750 + (7-distPawn)*4;
        return true;
    }
    int distFil = File( isqKing ) - File( isqPawn );
    if( distFil < 0 )
        distFil = - distFil;
    if( distFil >  (distPawn + myturn) ) {
        val = 750 + (7-distPawn)*4;
        return true;
    }
    return false;
}

/*
static bool SquareRuleKPK(PCON pcon, PSTE pste, int co, int & val) {
    // if dist to promote < dist from king then win
    bitboard pawns = pcon->pieces[PAWN|co];
    if(pawns) {
        int isqPawn = PopLSB(&pawns);
        return SquareRuleForPawn(pcon, pste, co, isqPawn, val);
    }
    return false;
}
*/

static bitboard BitIdx(int rank, int file) {
    if( rank >= 0 && rank <= 7 && file >=0 && file <= 7 )
        return IdxToU64( Idx(rank, file) );
    else
        return 0;
}

static inline bitboard BitIdx2(int rank, int file) {
    return IdxToU64( Idx(rank, file) );
}

static void inline mobDebugADD(int pc, int m) {
#ifdef _DEBUG
    mobDebugTable[mobDebugIdx].piece = pc;
    mobDebugTable[mobDebugIdx].mobility = m;
    mobDebugIdx++;
#endif
}


// friendly pawns in front of the king
// ennemy pawns in front of the king
// open files on the king
static int pawnOnKing(const CON *pcon, const STE *pste, int co, int sqr) {
    const U8 *pos = pcon->pos;
    int val = 0;
    int dir = co == W ? 1 : -1;
    int r = Rank(sqr);
    int f = File(sqr);
    bitboard myPawns = pcon->pieces[PAWN|co];
    bitboard theirPawns = pcon->pieces[PAWN|co^1];
    // the quality of the pawn shield depends on the number of attacker
    int a = (c_kingAtk[co^1] + 1) >> 1;

    if( (r+dir*3>=0 && r+dir*3<=7) )
    for(int ff = -1; ff <= 1; ff++) {
        int fp = f + ff;
        if( fp >=0 && fp <= 7 ) {
            //phashp->side[co^1].pawnatk
            bitboard thisPawn;
            if( (thisPawn = BitIdx2(r, fp)) & myPawns ) {
                val += 4; //6+a; //8;
            } else if( (thisPawn = BitIdx2(r + dir, fp)) & myPawns ) {
                val += 6; //6+a; //8;
//                if( thisPawn & phashp->side[co^1].pawnatk )
//                    val -= 4+(a>>1);
            } else if( (thisPawn = BitIdx2(r + dir*2, fp)) & myPawns ) {
                val += 3; //3+a; //3;
//                if( thisPawn & phashp->side[co^1].pawnatk )
//                    val -= 2+(a>>1);
            } else if( (thisPawn = BitIdx2(r + dir*3, fp)) & myPawns ) {
                val += -a; //0
//                if( thisPawn & phashp->side[co^1].pawnatk )
//                    val -= (a>>1);
            } else {
#if 1
                val -= 6+a+a;
#else
                // open file on rook and file sliders around
                if( kingAtkFSlider[co^1]) {
                    // no pawn to protect
                    if( phashp->side[co].c_pawn[fp] == 0 )
                        val -= 10+a;
//                        val -= 7+a+a;
                }
#endif
            }
            if( BitIdx2(r + dir, fp) & theirPawns )
                val -= 40+a; //50;
            else if( BitIdx2(r + dir*2, fp) & theirPawns )
                val -= 20+a; //35;
            else if( BitIdx2(r + dir*3, fp) & theirPawns )
                val -= 7+a; //15
            else {
                // open file on rook and file sliders around
                if( kingAtkFSlider[co^1]) {
                    // no ennemy pawns on this file ?
                    if( phashp->side[co^1].c_pawn[fp] == 0 )
                        val -= 12+a+a;
                }
            }
        }
    }
    return val;
}

// this has mostly to do with king safety
static int evalKingCastle(const CON *pcon, const STE *pste, int co) {
    int sqr = pcon->king[co];
    const U8 *pos = pcon->pos;
    int dir = co == W ? 1 : -1;
    int r = Rank(sqr);
    int f = File(sqr);
    // friendly pawns in front of the king
    // ennemy pawns in front of the king
    // compute safety value of the king present position
    int val = pawnOnKing(pcon, pste, co, sqr);
    // compute safety value of the king future position if the castle can still be done
    // if the king has not castled yet (perhaps because of some tactical moves)
    // and can still castle then blend the 2 scores

    int CastleSqrVal = -999;
    if( (co == W && (pste->castle & cfE1G1)) || (co == B && (pste->castle & cfE8G8))) {
        // king side castle
        CastleSqrVal = pawnOnKing(pcon, pste, co, co == W ? G1 : G8);
        // pieces between king and rook are delaying the castle
        if( pos[sqr+1] != 0 )   CastleSqrVal -= 7;
        if( pos[sqr+2] != 0 )   CastleSqrVal -= 7;
        val -= 5;
    }
    if( (co == W && (pste->castle & cfE1C1)) || (co == B && (pste->castle & cfE8C8))) {
        // queen side castle
        int QueenCastleSqrVal = pawnOnKing(pcon, pste, co, co == W ? C1 : C8);
        // pieces between king and rook are delaying the castle
        if( pos[sqr-1] != 0 )   QueenCastleSqrVal -= 5;
        if( pos[sqr-2] != 0 )   QueenCastleSqrVal -= 5;
        if( pos[sqr-3] != 0 )   QueenCastleSqrVal -= 5;
        if( QueenCastleSqrVal > CastleSqrVal )
            CastleSqrVal = QueenCastleSqrVal;
        val -= 5;
    }
    if( CastleSqrVal > val ) {
        // We can still castle and it's better for us, blend the scores.
        // Scores are averaged because the current score is not permanent.
        // how much of each ???
        val = (CastleSqrVal + val) / 2;
    }
//    if( pcon->pieces[QUEEN|co^1] == 0 )
//        val /= 2;
    return val;
}

    // for each piece of the board...
#define pieceheader(kind) \
		int sqr = PopLSB(&pieces);  \
		int pc = kind|co;          \
        int neutralPc = kind;       \
        /* neutral color square for positional values */ \
        int nSqr = remapSq[co][sqr]; \
        /* basic positional value */ \
        posval[co] += staticPosVal[neutralPc >> 1][nSqr]; \
        /*tempoval[co] += tempoVal[neutralPc >> 1][nSqr];*/ \
        if( neutralPc != PAWN ) \
            notpawn[co]++;


static INLINE void evalPAWN(const CON *pcon, const STE *pste, int co) {
    if( attackMap[co] & kingRing[co^1] )
        c_kingAtk[co] += ATK_PAWN;
    const U8 *pos = pcon->pos;
	for (bitboard pieces = pcon->pieces[PAWN|co]; pieces; ) {
        pieceheader(PAWN);

        posval[co] += pawnFileValMG[ File(sqr) ];
        endgval[co] += pawnFileValEG[ File(sqr) ];

        bitboard thispawn = IdxToU64(sqr);
        int rankBonus = pawn_bonus_to_7[co][Rank(sqr)];
        // give a little bonus for candidate pawns
        if( phashp->side[co].candidate & thispawn ) {
            // protected candidate
            if( thispawn & attackMap[co] )
                otherval[co] += 3;
//              posval[co] += 5 + rankBonus / 4;
            posval[co] += 5 + rankBonus / 2;
            endgval[co] += (5+rankBonus);
            endgval[co] += 2*sqDistance(sqr, pcon->king[co^1]);
            // this will incite the king to move to own candidate/passed pawn
            endgval[co] -= 1*sqDistance(sqr, pcon->king[co]);
        }
    }
}

static INLINE void evalPAWNpasser(const CON *pcon, const STE *pste, int co) {
    // we need the attack map for this evaluation
    const U8 *pos = pcon->pos;
	for (bitboard pieces = phashp->side[co].passed; pieces; ) {
		int sqr = PopLSB(&pieces);
		int pc = PAWN|co;
        int neutralPc = PAWN;
        int nSqr = remapSq[co][sqr];

        bitboard thispawn = IdxToU64(sqr);
        int rankBonus = pawn_bonus_to_7[co][Rank(sqr)];
        posval[co] += rankBonus;
        endgval[co] += (rankBonus+rankBonus);
        // give a little bonus for passed pawns
        // protected passed pawn
        int dir = -8;
        if( co == W )
            dir = 8;
        if( pos[sqr-1] == (PAWN|co) || pos[sqr+1] == (PAWN|co) 
            || pos[sqr-dir-1] == (PAWN|co) || pos[sqr-dir+1] == (PAWN|co)) {
            posval[co] += rankBonus;
            endgval[co]  += rankBonus + rankBonus;
        }
        // From Ed
        int count = 0;
        int s = sqr + dir;
        for(int i = 0; i < 3; i ++) {
            bitboard thisSqr = IdxToU64(s);
            if( pos[s] != 0 ) {
                if( ((pos[s] & 1) != co) && (attackMap[co] & thisSqr) )
                    count++;
                break;
            }
            if( (attackMap[co^1] & thisSqr) && ((attackMap[co] & thisSqr) == 0) )
                break;
            if( s < A2 || s > H7 ) {
                count = 3;
                break;
            }
            count++;
            s += dir;
        }
        posval[co] += count*rankBonus;
        endgval[co]  += count*(rankBonus + rankBonus);
        if( thispawn & king_attacks[ pcon->king[co] ] )
            endgval[co]  += rankBonus;
        // this will incite the king to move to candidate/passed pawn of other side
        endgval[co] += 4*sqDistance(sqr, pcon->king[co^1]);
        // this will incite the king to move to own candidate/passed pawn
        endgval[co] -= 2*sqDistance(sqr, pcon->king[co]);
        if( mat[co ^ 1] == 0 ) {
            int specialVal;
            // compute square rule
            SquareRuleForPawn(pcon, pste, co, sqr, specialVal);
            if( specialVal > 0 ) {
                endgval[co] += (specialVal);
            }
        }
    }
}

static INLINE void storeAttackMap(int co, int sqr, bitboard attacks, U8 type) {
    Assert( c_piece_attacks[co] < 16 );
    piece_attacks[co][ c_piece_attacks[co] ] = attacks;
    piece_list[co][ c_piece_attacks[co] ] = IdxToU64( sqr );
    piece_list_sqr[co][ c_piece_attacks[co] ] = sqr;
    piece_list_type[co][c_piece_attacks[co] ] = type;
    c_piece_attacks[co]++;
}

static INLINE void getAttackMap(int co, bitboard &thisPiece, bitboard &attacks, int &sqr) {
    Assert( c_piece_mob[co] < 16 );
    attacks = piece_attacks[co][ c_piece_mob[co] ];
    thisPiece = piece_list[co][ c_piece_mob[co] ];
    sqr = piece_list_sqr[co][ c_piece_mob[co] ];
    c_piece_mob[co]++;
}

static INLINE void evalKNIGHT(const CON *pcon, const STE *pste, int co) {
    const U8 *pos = pcon->pos;
	for (bitboard pieces = pcon->pieces[KNIGHT|co]; pieces; ) {
        pieceheader(KNIGHT);

        //	Computers are by default very stupid about putting
        //	bishops and knights on d3 and e3 while there is still
        //	a pawn on d2 or e2.  I will detect this case and
        //	penalize heavily.
        //
        if(co == W && (sqr == E3 || sqr == D3) && (pos[sqr - 8] == WP) )
            posval[co] -= 30;
        if(co == B && (sqr == E6 || sqr == D6) && (pos[sqr + 8] == BP) )
            posval[co] -= 30;

        bitboard moves = knight_attacks[sqr];
        tmp_attack[co] |= moves;
        storeAttackMap(co, sqr, moves, KNIGHT);
        if( moves & kingRing[co^1] )
            c_kingAtk[co] += ATK_MINOR;
        if( moves & kingRing[co] )
            c_kingDef[co] += ATK_MINOR;
        c_attack[co] += lowpopCount(moves & targets[co]);
//        bitboard futurOutpost = moves & outpostMap & attackMap[co] & ~attackMap[co^1];
//        int futurOutpostVal = 0;
//        if( futurOutpost ) {
//            futurOutpostVal = 10/2; // outpostValue[nSqr]/2;
//        }
//        int ov = evalKnightOutpost(sqr,co); 
        if( nSqr == B8 || nSqr == G8)
            underdev[co]++;
    }
}

static INLINE void evalKNIGHTmob(const CON *pcon, const STE *pste, int co) {
    int sqr;
    bitboard moves;
    bitboard thisPiece;
    getAttackMap( co, thisPiece, moves, sqr );
    moves &= ~pcon->pieces[B_OCCU|co];
//    int mob = (1+popCount(moves) * 2 - popCount(moves & attackMap[co^1]))>>1;
    int mob = popCount(moves);
    if( busy_pieces & thisPiece )
        mob = (mob + 1) >> 1;
    mobDebugADD(KNIGHT|co,mob);
    c_mobility[co] += mob;
    Assert(mob>=0 && mob <= 8);
    mobility[co] += KnightMobility[mob];
    // knight outpost
    int nSqr = remapSq[co][sqr];
    int bonus = knightOutpostVal[nSqr];
    if( bonus ) {
        bonus = 1 + (bonus >> 1);
        // not attacked by an ennemy pawn
        if( (thisPiece & phashp->side[co^1].pawnatk) == 0 )
            posval[co] += bonus;
            // defended by my pawns
            if( thisPiece & phashp->side[co^1].pawnatk )
                posval[co] += bonus;
            // not attacked at all
            if( (thisPiece & attackMap[co^1]) == 0 )
                posval[co] += bonus;
    }

}

static INLINE void evalBISHOP(const CON *pcon, const STE *pste, int co) {
    const U8 *pos = pcon->pos;
    // don't reduce the mobility of linked bishop/queen
    bitboard bishopBlocker = pcon->pieces[OCCU] ^ pcon->pieces[QUEEN|co];
	for (bitboard pieces = pcon->pieces[BISHOP|co]; pieces; ) {
        pieceheader(BISHOP);
        bishopSquare[co] = sqr;
        //	Computers are by default very stupid about putting
        //	bishops and knights on d3 and e3 while there is still
        //	a pawn on d2 or e2.  I will detect this case and
        //	penalize heavily.
        //
        if(co == W && (sqr == E3 || sqr == D3) && (pos[sqr - 8] == WP) )
            posval[co] -= 30;
        if(co == B && (sqr == E6 || sqr == D6) && (pos[sqr + 8] == BP) )
            posval[co] -= 30;

        bitboard moves = Bmagic(sqr, bishopBlocker);
        tmp_attack[co] |= moves;
        storeAttackMap(co, sqr, moves, BISHOP);
        c_attack[co] += lowpopCount(moves & targets[co]);

        moves = Bmagic(sqr, moveBlocker[co]);
        if( moves & kingRing[co^1] )
            c_kingAtk[co] += ATK_MINOR;
        if( moves & kingRing[co] )
            c_kingDef[co] += ATK_MINOR;

        // the bishop is a little better than the knight in eg
        endgval[co] += 25;
        if( nSqr == C8 || nSqr == F8)
            underdev[co]++;
    }
    if( pcon->msigCount[BISHOP|co] >= 2 ) {
        // bishop bonus if open position
        posval[co] += (6 - phashp->c_center_pawns);
        // bishop pair
        posval[co] += 45;
    }

}

static INLINE void evalBISHOPmob(const CON *pcon, const STE *pste, int co) {
    int sqr;
    bitboard moves;
    bitboard thisPiece;
    getAttackMap( co, thisPiece, moves, sqr );
    moves &= ~pcon->pieces[B_OCCU|co];
//    int mob = (1+popCount(moves) * 2 - popCount(moves & attackMap[co^1]))>>1;
    int mob = popCount(moves);
    if( busy_pieces & thisPiece )
        mob = (mob + 1) >> 1;
    mobDebugADD(BISHOP|co,mob);
    c_mobility[co] += mob;
    Assert(mob>=0 && mob <= 14);
    mobility[co] += BishopMobility[mob];
    // bishop outpost
    int nSqr = remapSq[co][sqr];
    int bonus = knightOutpostVal[nSqr];
    if( bonus ) {
        bonus = 1 + (bonus >> 1);
        // not attacked by an ennemy pawn
        if( (thisPiece & phashp->side[co^1].pawnatk) == 0 )
            posval[co] += bonus;
            // defended by my pawns
            if( thisPiece & phashp->side[co^1].pawnatk )
                posval[co] += bonus;
            // not attacked at all
            if( (thisPiece & attackMap[co^1]) == 0 )
                posval[co] += bonus;
    }
}

static INLINE void evalROOK(const CON *pcon, const STE *pste, int co) {
    const U8 *pos = pcon->pos;
    // don't reduce the mobility of linked rooks
    bitboard rookBlocker = pcon->pieces[OCCU] ^ pcon->pieces[ROOK|co] ^ pcon->pieces[QUEEN|co];
	for (bitboard pieces = pcon->pieces[ROOK|co]; pieces; ) {
        pieceheader(ROOK);

        bitboard moves = Rmagic(sqr, rookBlocker);
        // no Xray for def
        if( moves & kingRing[co] )
            c_kingDef[co] += ATK_ROOK;
        tmp_attack[co] |= moves;
        storeAttackMap(co, sqr, moves, ROOK);
        c_attack[co] += lowpopCount(moves & targets[co]);

        int rookF = File(sqr);
        // Xray attack
        moves = Rmagic(sqr, moveBlocker[co]);
        if( moves & kingRing[co^1] ) {
            c_kingAtk[co] += ATK_ROOK;
            kingAtkFSlider[co] = true;
            if( rookF == filH && (phashp->side[co^1].c_pawn[rookF] == 0))
                c_kingAtk[co]++;
        }

        // rook on semi-open file
        if(phashp->side[co].c_pawn[rookF] == 0) {
            posval[co] += 7;
            // rook on open file
            if(phashp->side[co^1].c_pawn[rookF] == 0)
                posval[co] += 7;
        }
        //
        endgval[co] += 75;
        if( pste->valPcThem < valROOK )
            endgval[co] += 75;
        int knSqr = remapSq[co][pcon->king[co]];
        // trapped rook
        if( phashp->side[co].c_pawn[rookF] ) {
            if( nSqr == A8 && (knSqr == B8 || knSqr == C8) )
                posval[co] -= 50;
            if( nSqr == B8 && (knSqr == C8) )
                posval[co] -= 50;
            if( nSqr == H8 && (knSqr == F8 || knSqr == G8) )
                posval[co] -= 50;
            if( nSqr == G8 && (knSqr == F8) )
                posval[co] -= 50;
        }
    }
}

static INLINE void evalROOKmob(const CON *pcon, const STE *pste, int co) {
    int sqr;
    bitboard moves;
    bitboard thisPiece;
    getAttackMap( co, thisPiece, moves, sqr );

    moves &= ~pcon->pieces[B_OCCU|co];
    // counts only half the mobility for horizontal moves
    bitboard horMoves = moves & rankMask[ Rank(sqr) ];
//    int mob = (popCount(moves) * 2 - popCount(moves & attackMap[co^1]))>>1;
//    int mob = (1+popCount(moves) * 2 - popCount(horMoves | (moves & attackMap[co^1])))>>1;
    int mob = (1+popCount(moves) * 2 - popCount(horMoves))>>1;
    if( busy_pieces & thisPiece )
        mob = (mob + 1) >> 1;
    mobDebugADD(ROOK|co,mob);
    c_mobility[co] += mob;
    Assert(mob>=0 && mob <= 14);
    mobility[co] += RookMobility[mob];
}


static INLINE void evalQUEEN(const CON *pcon, const STE *pste, int co) {
    const U8 *pos = pcon->pos;
	for (bitboard pieces = pcon->pieces[QUEEN|co]; pieces; ) {
        pieceheader(QUEEN);
        if(co == W && (sqr == E3 || sqr == D3) && (pos[sqr - 8] == WP) )
            posval[co] -= 20;
        if(co == B && (sqr == E6 || sqr == D6) && (pos[sqr + 8] == BP) )
            posval[co] -= 20;

//        bitboard moves = Qmagic(sqr, pcon->pieces[OCCU]);
        bitboard moves = Rmagic(sqr, pcon->pieces[OCCU]^pcon->pieces[ROOK|co]) |
            Bmagic(sqr, pcon->pieces[OCCU]^pcon->pieces[BISHOP|co]);
        tmp_attack[co] |= moves;
        storeAttackMap(co, sqr, moves, QUEEN);
        c_attack[co] += lowpopCount(moves & targets[co]);

        moves = Qmagic(sqr, moveBlocker[co]);
        if( moves & kingRing[co^1] ) {
            c_kingAtk[co] += ATK_QUEEN;
            kingAtkFSlider[co] = true;
        }
        if( moves & kingRing[co] )
            c_kingDef[co] += ATK_QUEEN;

        // don't move the queen before the minors
        if( (nSqr != D8) && (notpawn[co] >= 5) ) {
            posval[co] -= underdev[co] * 8;
        }
        endgval[co] += 75;
        // queen counts for two
        notpawn[co]++;
    }
}

static INLINE void evalQUEENmob(const CON *pcon, const STE *pste, int co) {
    int sqr;
    bitboard moves;
    bitboard thisPiece;
    getAttackMap( co, thisPiece, moves, sqr );
    moves &= ~pcon->pieces[B_OCCU|co];
//    int mob = (1+popCount(moves) * 2 - popCount(moves & attackMap[co^1]))>>1;
    int mob = popCount(moves);
    if( busy_pieces & thisPiece )
        mob = (mob + 1) >> 1;
    mobDebugADD(QUEEN|co,mob);
    c_mobility[co] += mob;
    Assert(mob>=0 && mob <= 28);
    mobility[co] += QueenMobility[mob];
}

static INLINE void evalKING(const CON *pcon, const STE *pste, int co) {
    const U8 *pos = pcon->pos;
    int sqr = pcon->king[co];
    int neutralPc = KING;
    int pc = KING|co;
    /* neutral color square for positional values */
    int nSqr = remapSq[co][sqr];
    /* basic positional value */
    posval[co] += staticPosVal[neutralPc >> 1][nSqr];
    notpawn[co]++;

    if( mat[co ^ 1] > 1100 )
        safetyval[co] += evalKingCastle(pcon, pste, co);

    endgval[co] += KingEGPosVal[nSqr];
    // add king mobility for endgame
    bitboard moves = king_attacks[sqr];
    storeAttackMap(co, sqr, moves, KING);
    moves &= ~attackMap[co^1];
    tmp_attack[co] |= moves;
    moves &= ~pcon->pieces[B_OCCU|co];
    int mob = popCount(moves);
    mobDebugADD(pc,mob);
    endgval[co] += mob*2;
    if( true && notpawn[co] <= 3 ) {
        // accessible squares
        bitboard kingSqr = IdxToU64(sqr);
        bitboard myStuff = pcon->pieces[B_OCCU|co];
        // acessible squares : not occupied by my pawns/pieces, not attacked by their pawns/pieces
        bitboard path = 0xffffffffffffffffULL & ((~myStuff)|kingSqr) & ~attackMap[co ^ 1];
        // look for pawns not defended by pawns
//        bitboard pieces = pcon->pieces[PAWN|co^1] & ~phashp->side[co^1].pawnatk;
        bitboard pieces = pcon->pieces[PAWN|co^1] & ~phashp->side[co^1].pawnatk & ~king_attacks[pcon->king[co^1]];
        pawnDist[co] = 99;
	    if ( pieces ) {
            int dist = pathLength(kingSqr, pieces, path);
            pawnDist[co] = dist;
        }
        // blocked position, for at least one side
        if( pawnDist[co] > 10 ) {
            if( notpawn[co] == 1 ) {
                endgval[co] /= 2;
                otherval[co] /= 2;
                posval[co] /= 2;
            } else
                endgval[co] -= (10)*(9-notpawn[co]-notpawn[co]);
//            endgval[co] = (3*endgval[co]) / 4;
//            otherval[co] /= 2;
        } else
            endgval[co] -= (pawnDist[co]-1)*(9-notpawn[co]-notpawn[co]);
    }

}

//	This eval function evaluates:
//
//	1.	Pawn structure, which is hashed and handled in another module.
//	2.	Piece location, which is usually related to central occupation or
//		in the case of pawns, simple advancement.
//	3.	Minors blocking d2/e2 pawns.
//	4.	King on f1/g1 with rook on g1/h1.
//	5.	Pawn shelter.
//	6.	Endgame king position values versus middlegame values.
//
//	Cases 3 and 4 are surprisingly important.
//
//	The piece-square tables are designed to induce the program to castle
//	behind a safe pawn structure, to move its central pawns, and to develop
//	its pieces.
//
//	There is also a chance that the program would feel an interest in breaking
//	its opponent's king position, but I don't have high hopes for this.

static void dbgPos(char *txt) {
    printf("# W:pos=%d endg=%d mob=%d B:pos=%d endg=%d mob=%d \t(after %s)\n", 
        posval[W], endgval[W], mobility[W], posval[B], endgval[B], mobility[B], txt);
}
#ifdef _DEBUG
#define DBGPOS(txt) if( Alpha == -20000 ) dbgPos(txt)
#else
#define DBGPOS(txt)
#endif



int ValEval(PCON pcon, PSTE pste, int Alpha, int Beta)
{
    if( pste->evaluated && !pste->evalislazy )
        return pste->val;

    int specialVal;
    bool isPawnEndGame = false;
    int side = pste->side;
    int xside = side ^ 1;
    int isqKingUs = pcon->king[side];
    int isqKingThem = pcon->king[xside];
    mobDebugIdx = 0;
    stats.cEval++;
    pste->bound = fail_bound;

    const EvalCacheEntry *thisEval = evalCache->find( pste );
    if( thisEval ) {
        stats.evc_hit++;
        stats.cFullEval++;
        pste->evaluated = true;
        pste->evalislazy = false;
        pste->val = thisEval->val;
        return thisEval->val;
    }
    // -- probe bitbases --
    // handle only 4 men
    if( true && (pcon->msigCount[WP] + pcon->msigCount[BP]) <= 2 )
        if((pcon->msigCount[WP] + pcon->msigCount[BP] +
            pcon->msigCount[WN] + pcon->msigCount[BN] + 
            pcon->msigCount[WB] + pcon->msigCount[BB] +
            pcon->msigCount[WR] + pcon->msigCount[BR] +
            pcon->msigCount[WQ] + pcon->msigCount[BQ]) <= 2 ) {
            // do probe...
            if( probe_bitbases(pcon, pste, side, specialVal) ) {
                stats.c_recog++;
                pste->evaluated = true;
                pste->evalislazy = false;
                pste->val = specialVal;
                pste->bound = exact_bound;
                return specialVal;
            }
        }

    // -- call recognizers --
    unsigned b_mstate = msigToIdx( pcon->msigBit[B] );
    unsigned w_mstate = msigToIdx( pcon->msigBit[W] );
    if( true && recog_available(b_mstate,w_mstate)) {
        recog thisRecog = recog_table[recog_index(w_mstate,b_mstate)];
        recog_result ret = {fail_bound,0,recog_need_validation};
        if( thisRecog(pcon, pste, ret) ) {
            // we can not return a bound for some positions, convert the bound into a simple score
            if( (ret.recog_validation == recog_need_validation) && (ret.bound != score_bound) 
                && !positionOkForRecog(pcon, pste) )
                ret.bound = score_bound;
            stats.c_recog++;
            pste->evaluated = true;
            pste->evalislazy = false;
            pste->val = ret.score;
            pste->bound = int(ret.bound);
            return ret.score;
        }
    }
    pste->evalislazy = true;

	//	Get value for pawn structure.
	//
    EvalPawns(pcon, pste, phashp);
	//
	//	Get material value.
	//
    int val[coMAX];
    // pawn structure score not included in eg, structure is added in posval only
    val[side] = pste->valPcUs + pste->valPnUs + phashp->side[side].val ;
    val[xside] = pste->valPcThem + pste->valPnThem + phashp->side[xside].val ;
//    val[side] = pste->valPcUs + pste->valPnUs;
//    val[xside] = pste->valPcThem + pste->valPnThem;
    mat[side] = pste->valPcUs;
    mat[xside] = pste->valPcThem;

    /////
    /////return val[side] - val[xside];
    /////

    //
    //  Check for insuffisant material
    //
    // No pawn
	if ( (pste->valPnUs + pste->valPnThem) == 0 ) {
        // K vs K
		if ( (pste->valPcUs + pste->valPcThem)== 0 )
            return 0;
        // K + Minor vs K
		if ((pste->valPcUs == valBISHOP || pste->valPcUs == valKNIGHT) && (pste->valPcThem == 0))
			return 0;
		if ((pste->valPcUs == 0) && (pste->valPcThem == valBISHOP || pste->valPcThem == valKNIGHT))
			return 0;
        // Two knights can not force checkmate
        if ((pste->valPcUs == 2*valKNIGHT)) {
            if (pste->valPcThem == 0 || pste->valPcThem == valKNIGHT || pste->valPcThem == valBISHOP)
			    return 50;
        }
        if ((pste->valPcThem == 2*valKNIGHT)) {
            if (pste->valPcUs == 0 || pste->valPcUs == valKNIGHT || pste->valPcUs == valBISHOP)
    			return -50;
        }
        // K + Q vs K
		if ((pste->valPcUs == valQUEEN||pste->valPcUs == valROOK) && (pste->valPcThem == 0))
			return pste->valPcUs + 150 - c_argdistborder[ isqKingThem ] -
                sqDistance( isqKingUs, isqKingThem );
        // K vs K + Q | R
		if ((pste->valPcUs == 0) && (pste->valPcThem == valQUEEN||pste->valPcThem == valROOK))
			return -(pste->valPcThem + 150 - c_argdistborder[ isqKingUs ] -
                sqDistance( isqKingUs, isqKingThem ));
        if (pste->valPcUs == (valBISHOP + valKNIGHT)) {
            if( (pste->valPcThem == valBISHOP||pste->valPcThem == valKNIGHT))
                return 50;
            // KBNK
            else if( (pste->valPcThem == 0))
                if( KBNK( pcon, pste, side, specialVal) )
                    return specialVal;
        }
        if (pste->valPcThem == (valBISHOP + valKNIGHT)) {
            if( (pste->valPcUs == valBISHOP||pste->valPcUs == valKNIGHT))
                return -50;
            else if( KBNK( pcon, pste, xside, specialVal) )
                return -specialVal;
        }
    }
    // No piece
    if ((pste->valPcUs + pste->valPcThem)== 0) {
        isPawnEndGame = true;
        // only a pawn on one side
    	if ( (pste->valPnUs == valPAWN) && (pste->valPnThem == 0) ) {
            // compute ranks from pawn to promotion
            // compute file dist from opposite king to pawn
            // if dist to promote < dist from king then win
            if( KPK( pcon, pste, side, specialVal) )
                return specialVal;
        }
    	if ( (pste->valPnUs == 0) && (pste->valPnThem == valPAWN) ) {
            // compute ranks from pawn to promotion
            // compute file dist from opposite king to pawn
            // if dist to promote < dist from king then win
            if( KPK( pcon, pste, xside, specialVal) )
                return -specialVal;
        }
    }
    // bishop + pawns vs nothing
    if ((pste->valPcUs == valBISHOP || pste->valPcUs == 0) && (pste->valPcThem == 0) &&
        (pste->valPnUs > 0) && (pste->valPnThem == 0)) {
        // pawns on file A
        if( (pcon->pieces[PAWN|side] & ~fileAMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, side, promotionSqrFileA[side]) && sqDistance( pcon->king[xside], promotionSqrFileA[side] ) <= 1 )
                return 50;
        // pawns on file H
        if( (pcon->pieces[PAWN|side] & ~fileHMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, side, promotionSqrFileH[side]) && sqDistance( pcon->king[xside], promotionSqrFileH[side] ) <= 1 )
                return 50;
    }
    // bishop + pawns vs nothing
    if ((pste->valPcThem == valBISHOP || pste->valPcThem == 0) && (pste->valPcUs == 0) &&
        (pste->valPnThem > 0) && (pste->valPnUs == 0)) {
        // pawns on file A
        if( (pcon->pieces[PAWN|xside] & ~fileAMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, xside, promotionSqrFileA[xside]) && sqDistance( pcon->king[side], promotionSqrFileA[xside] ) <= 1 )
                return -50;
        // pawns on file H
        if( (pcon->pieces[PAWN|xside] & ~fileHMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, xside, promotionSqrFileH[xside]) && sqDistance( pcon->king[side], promotionSqrFileH[xside] ) <= 1 )
                return -50;
    }

    //
    //	Cut off if it appears that this is futile.
    //  Don't do this cut off in endgames !!
    //
#if 0
    // caution with score and pruning
    int tmpval = val[side] - val[xside];
    int margin = 900;
    const int posMargin = 50;
    if( !isPawnEndGame )
        if( (Alpha+1) == Beta )
            margin = 700;
	if (tmpval - margin >= Beta)		// We're already great?
        return (tmpval - posMargin) ? (tmpval - posMargin) : 1;
//	if (tmpval + margin <= Alpha) 	// We're already toast?
//        return (tmpval + posMargin) ? (tmpval + posMargin) : -1;
#endif

    attackMap[B] = attackMap[W] = 0;
    posval[B] = posval[W] = 0;      // position value
    tacval[B] = tacval[W] = 0;      // tactical value
    otherval[B] = otherval[W] = 0;  // other value ?
    endgval[B] = endgval[W] = 0;    // end game values
    notpawn[B] = notpawn[W] = 0;
    attackval[B] = attackval[W] = 0;
    c_mobility[B] = c_mobility[W] = 0;
    mobility[B] = mobility[W] = 0;
    defSpace[B] = defSpace[W] = 0;
    pawnDist[B] = pawnDist[W] = 99;
    c_attack[B] = c_attack[W] = 0;
    kingRing[B] = king_attacks[pcon->king[B]];
    kingRing[B] |= soutOne(kingRing[B]);
    kingRing[W] = king_attacks[pcon->king[W]];
    kingRing[W] |= nortOne(kingRing[W]);
    // for xray king attacks
    moveBlocker[B] = pcon->pieces[B_OCCU|W] | pcon->pieces[PAWN|B] | pcon->pieces[KING|B];
    moveBlocker[W] = pcon->pieces[B_OCCU|B] | pcon->pieces[PAWN|W] | pcon->pieces[KING|W];
    // for normal attacks
    targets[B] = pcon->pieces[B_OCCU|W] ^ (pcon->pieces[PAWN|W] & phashp->side[W].pawnatk);
    targets[W] = pcon->pieces[B_OCCU|B] ^ (pcon->pieces[PAWN|B] & phashp->side[B].pawnatk);
    // king tropism, attack & defense
    c_kingAtk[B] = c_kingAtk[W] = 0;
    c_kingDef[B] = c_kingDef[W] = 0;
    safetyval[B] = safetyval[W] = 0;
    kingAtkFSlider[B] = kingAtkFSlider[W] = false;
    c_piece_attacks[B] = c_piece_attacks[W] = 0;
    c_piece_mob[B] = c_piece_mob[W] = 0;
    attacked_material[B] = attacked_material[W] = 0;
    busy_pieces = 0;

    const bitboard offenseMapSide[coMAX] = {0x00000000ffffffffULL, 0xffffffff00000000ULL};
//    posval[W] = phashp->side[W].val;
//    posval[B] = phashp->side[B].val;
//    otherval[W] = phashp->side[W].val;
//    otherval[B] = phashp->side[B].val;

    // hack alert : change the content of the king square
    // still needed for pawn passer eval
    pcon->pos[pcon->king[B]] = APC_BLACKKING;

    const U8 *pos = pcon->pos;

    {
	    // - Pawns seven-shift -
	    attackMap[W] |= ((pcon->pieces[PAWN|W] & CLEAR_LEFT) << 7);
	    // - Pawns nine-shift -
	    attackMap[W] |= ((pcon->pieces[PAWN|W] & CLEAR_RIGHT) << 9);
	    // - Pawns seven-shift -
	    attackMap[B] |= ((pcon->pieces[PAWN|B] & CLEAR_RIGHT) >> 7);
	    // - Pawns nine-shift -
	    attackMap[B] |= ((pcon->pieces[PAWN|B] & CLEAR_LEFT) >> 9);
    }

#define foreachside2(c,code)    {int co = c; code;}
#define foreachside(code)       foreachside2(B,code) foreachside2(W,code)

    DBGPOS("init");
    evalPAWN(pcon, pste, B);
    evalPAWN(pcon, pste, W);
    DBGPOS("pawn");

    foreachside(
        defSpace[co] = phashp->side[co].c_space;
        if( File(pcon->king[co]) >= filD )
            defSpace[co] += phashp->side[co].c_Qspace;
        else if( File(pcon->king[co]) <= filE )
            defSpace[co] += 2*phashp->side[co].c_Kspace;
    )
    posval[B] += defSpace[B];
    posval[W] += defSpace[W];
    DBGPOS("space");

    knight_moves[B] = knight_moves[W] = 0;
    tmp_attack[B] = tmp_attack[W] = 0;
    underdev[B] = underdev[W] = 0;

    evalKNIGHT(pcon, pste, B);
    evalKNIGHT(pcon, pste, W);
    DBGPOS("KNIGHT");

    bishopSquare[B] = bishopSquare[W] = 0;

    evalBISHOP(pcon, pste, B);
    evalBISHOP(pcon, pste, W);
    DBGPOS("BISHOP");

    bool drawish = false;
    // EG with pawns and bishops of opposite color
    if( (pste->valPcThem == valBISHOP) && (pste->valPcUs == valBISHOP) ) {
        if( abs(pste->valPnThem - pste->valPnUs) <= valPAWN * 2 )
            if( !sameSquareColor( bishopSquare[W], bishopSquare[B] ) )
                drawish = true;
    }
    attackMap[W] |= tmp_attack[W];
    attackMap[B] |= tmp_attack[B];
    tmp_attack[W] = tmp_attack[B] = 0;

    evalROOK(pcon, pste, B);
    evalROOK(pcon, pste, W);
    DBGPOS("ROOK");

    attackMap[W] |= tmp_attack[W];
    attackMap[B] |= tmp_attack[B];
    tmp_attack[W] = tmp_attack[B] = 0;

    evalQUEEN(pcon, pste, B);
    evalQUEEN(pcon, pste, W);
    DBGPOS("QUEEN");

    attackMap[W] |= tmp_attack[W];
    attackMap[B] |= tmp_attack[B];
    tmp_attack[W] = tmp_attack[B] = 0;

    evalKING(pcon, pste, B);
    evalKING(pcon, pste, W);
    DBGPOS("KING");

    attackMap[W] |= tmp_attack[W];
    attackMap[B] |= tmp_attack[B];


    // end of table
    piece_list_type[B][c_piece_attacks[B] ] = PIECES;
    piece_list_type[W][c_piece_attacks[W] ] = PIECES;

    foreachside(
        // all the stuff but the king, that is attacked, not defended by our pawns,
        // not defended by our king
//        attacked_material[co] = (pcon->pieces[B_OCCU|co] & ~pcon->pieces[KING|co]) &
//            attackMap[co^1] & (~phashp->side[co].pawnatk) & (~king_attacks[pcon->king[co]]);
//0.3b1   attacked_material[co] = (pcon->pieces[B_OCCU|co] & ~pcon->pieces[KING|co]) &
//            attackMap[co^1] & (~phashp->side[co].pawnatk);
        attacked_material[co] = (pcon->pieces[B_OCCU|co] & ~pcon->pieces[KING|co]) &
            attackMap[co^1];

    	for (bitboard pieces = attacked_material[co]; pieces; ) {
		    int sqr = PopLSB(&pieces);
            bitboard thisPiece = IdxToU64( sqr );
            bitboard tmp_def_list = 0;
            int c_def = 0;
            // how many times is it defended ? (pawns)
            if( sqr >= A2 && sqr <= H6 )
                if( co == W ) {
                    if( pos[sqr-7] == WP ) c_def++;
                    if( pos[sqr-9] == WP ) c_def++;
                } else {
                    if( pos[sqr+7] == BP ) c_def++;
                    if( pos[sqr+9] == BP ) c_def++;
                }
            // how many times is it defended ? (pieces)
            for(int i = c_piece_attacks[co]-1; i >= 0; i--) {
                if( piece_attacks[co][i] & thisPiece ) {
                    c_def++;
                    tmp_def_list |= piece_list[co][i];
                }
            }
            if( c_def <= 1 ) {  // common case
                // attacked more or equal number of times
                busy_pieces |= tmp_def_list;    // tmp_def_list can be empty here
            } else {
                // how many times is it attacked ? (pawns)
                if( sqr >= A2 && sqr <= H6 ) {
                    if( co == B ) {
                        if( pos[sqr-7] == WP ) c_def--;
                        if( pos[sqr-9] == WP ) c_def--;
                    } else {
                        if( pos[sqr+7] == BP ) c_def--;
                        if( pos[sqr+9] == BP ) c_def--;
                    }
                }
                // how many times is it attacked ? (pieces)
                for(int i = c_piece_attacks[co^1]-1; c_def > 0 && i >= 0; i--) {
                    if( piece_attacks[co^1][i] & thisPiece )
                        c_def--;
                }
                if( c_def <= 0 ) {
                    // attacked more or equal number of times
                    busy_pieces |= tmp_def_list;
                }
            }
        }

    )

    evalPAWNpasser(pcon, pste, B);
    evalPAWNpasser(pcon, pste, W);
    DBGPOS("pawn");

    foreachside(
        while( piece_list_type[co][ c_piece_mob[co] ] == KNIGHT ) {
            evalKNIGHTmob(pcon, pste, co);
        }
    )
    DBGPOS("KNIGHT mob");

    foreachside(
        while( piece_list_type[co][ c_piece_mob[co] ] == BISHOP ) {
            evalBISHOPmob(pcon, pste, co);
        }
    )
    DBGPOS("BISHOP mob");

    foreachside(
        while( piece_list_type[co][ c_piece_mob[co] ] == ROOK ) {
            evalROOKmob(pcon, pste, co);
        }
    )
    DBGPOS("ROOK mob");

    foreachside(
        while( piece_list_type[co][ c_piece_mob[co] ] == QUEEN ) {
            evalQUEENmob(pcon, pste, co);
        }
    )
    DBGPOS("QUEEN mob");

    foreachside(
        c_piece_mob[co]++;
    )
    Assert( c_piece_mob[W] == c_piece_attacks[W] );
    Assert( c_piece_mob[B] == c_piece_attacks[B] );

    posval[W] += phashp->side[W].c_center * notpawn[B];
    posval[B] += phashp->side[B].c_center * notpawn[W];
    DBGPOS("center");

    int medmob = (mobility[W] + mobility[B]) / 2;
#if 1
    foreachside(
        // -20 -10 0 10 20 30
        // -10 0  10 20 30 40
        if(true) {
            int mob = (mobility[co] - medmob + 30 + 0*c_attack[co]);
            if( mob > 0 ) {
                int activ = (c_attack[co] * (2+notpawn[co]) * mob) 
                    / 64;
    //                / 50;
    //                / 40;   // bad
                if( activ > 60 )
                    activ = 60;
                posval[co] += activ;
            }
            posval[co] += 4*c_attack[co];
        } else {
            posval[co] += c_attack[co] * 5;
        }
    )
    DBGPOS("targets");
#endif

    // pawns attacked and not defended by other pawns
    // gives a penality because we don't want to defend with a piece
    // with 10 points penality, the engine wants to attack with his queen
    // in the begining
    attackval[W] += 4*popCount( pcon->pieces[PAWN|B] & attackMap[W] & ~phashp->side[B].pawnatk );
    attackval[B] += 4*popCount( pcon->pieces[PAWN|W] & attackMap[B] & ~phashp->side[W].pawnatk );


    int valAttacking[coMAX] = {0,0};
    int valHanging[coMAX] = {0,0};

#define updateThreat(bm,val)   \
            if( bm & attackMap[co^1]) { \
                valAttacking[co ^ 1] ++; \
                if( (bm & attackMap[co]) == 0 ) \
                    valHanging[co] += val;\
                else if( bm & phashp->side[co^1].pawnatk ) \
                    valHanging[co] += val - 100;\
            }

#if 1
    if( true )  { // if not at leaf
        foreachside({
            updateThreat(pcon->pieces[PAWN|co],100);
            updateThreat(pcon->pieces[KNIGHT|co],325);
            updateThreat(pcon->pieces[BISHOP|co],325);
            updateThreat(pcon->pieces[ROOK|co],500);
            updateThreat(pcon->pieces[QUEEN|co],975);
        })
    } 
#endif

    // hack alert : restore the content of the king square
    pcon->pos[pcon->king[B]] = KING|B;

#ifdef KEYSQUARE
    if( isPawnEndGame ) {
        for(int i = 0; i < 3 ; i ++ )
            if( phashp->keySquare[i].square ) {
                int winner = phashp->keySquare[i].winningSide;
                int dist = phashp->keySquare[i].distance[winner^1][ pcon->king[winner^1] ] - 
                    phashp->keySquare[i].distance[winner][ pcon->king[winner] ];
                // side is side to move next
                if( dist >= 2 || (dist >= 1) && (winner == side) ) {
//                if( dist > 3 || (dist > 2) && (winner == side) ) {
                    endgval[winner] += dist * 10;
                }
            }
    }
#endif

//    int c_pawn = pcon->msigCount[PAWN|B] + pcon->msigCount[PAWN|W];

	for (int co = coBLACK; co <= coWHITE; co++) {
        // penalize for not developping minors (this is also a tempo counter in the begining)
        posval[co] -= underdev[co] * 20;
        // a*3-d
        // 4/4=>8 6/6=>12 7/7=>14 9/9=>18
        // 4/2=>10 6/4=>14 7/4=>17 9/2=>25
        // a*4-d
        // 4/4=>12 6/6=>18 7/7=>21 9/9=>27
        // 4/2=>14 6/4=>20 7/4=>24 9/2=>34
        // 2/4=>4 4/6=>10 0/7=>-7 2/9=>-1
        int a = c_kingAtk[co]*3 - c_kingDef[co^1];
//        int a = c_kingAtk[co]*4 - c_kingDef[co^1];
//        int a = c_kingAtk[co]*2;
        if( a > 0 )
            safetyval[co^1] -= a;

        // lower the value of minor pieces in endgame depending on the pawn count
        if( pcon->msigCount[BISHOP|co] + pcon->msigCount[KNIGHT|co] )
            val[co] -= minor_eg_malus[pcon->msigCount[PAWN|co]];

//        val[co] -= valHanging[co] / 50;
        // compute the mobility and position control
        // but only in midgame
        if( MidGameValueCoef[notpawn[ co ^ 1]] > 0 ) {
            // increase the mobility for the side with more material (pawns not included)
            if( ((mat[co] + 100) > mat[co^1]) ) //&& (mobility[co] > 0) )
                posval[co] += 50;
//                tacval[co] += (3*mobility[co]) / 3;
            tacval[co] += (2*mobility[co]) / 3;
        }
        if( EndGameValueCoef[notpawn[ co ^ 1]] > 0 ) {
            endgval[co] += mobility[co] / 2;
        }
        // knight bonus for closed positions
        if( pcon->msigCount[KNIGHT|co] >= 1 && pcon->msigCount[BISHOP|(co^1)] == 1 )
            posval[co] += (phashp->c_center_pawns - 6);
    }

    posval[W] += safetyval[W];
    posval[B] += safetyval[B];
    pste->kingsafety[W] = safetyval[W];
    pste->kingsafety[B] = safetyval[B];
    pste->threat[W] = valHanging[W];
    pste->threat[B] = valHanging[B];

    DBGPOS("...");

    int finalVal[coMAX];
    // xside is the side to move
//    val[xside] += 2;
//    val[side] += 10;
	finalVal[side] = val[side] + MidGameValue( posval[side] + tacval[side] + attackval[side],notpawn[xside]) + 
           EndGameValue( endgval[side], notpawn[xside] ) + otherval[side] + valAttacking[side];
	finalVal[xside] = val[xside] + MidGameValue( posval[xside] + tacval[xside] + attackval[xside],notpawn[side]) + 
           EndGameValue( endgval[xside], notpawn[side] ) + otherval[xside] + valAttacking[xside];

    int delta = finalVal[side] - finalVal[xside];
#if 0
    int sum = 400 + (((finalVal[side] + finalVal[xside]) > 0) ? (finalVal[side] + finalVal[xside]) : - (finalVal[side] + finalVal[xside]));
    int ratioBonus = delta * 128 / sum;
#else
    // pieces start:14=>0, mid:10=>4, eg:4=>10, eg:2=>12, eg:0=>14
    // pawns  start:16=>0, mid:12=>4, eg:8=>8, eq:4=>12
//    int exchange = 4*(14-(notpawn[W]+notpawn[B]))-8*(16-(c_pieces[PAWN|B] + c_pieces[PAWN|W]));
    int exchange = 4*(10-(notpawn[W]+notpawn[B]))-8*(10-(pcon->msigCount[PAWN|B] + pcon->msigCount[PAWN|W]));
//    int ratioBonus = exchange * delta / 256;
//    int ratioBonus = exchange * (val[side]-val[xside]) / 256;
    int ratioBonus = exchange * (finalVal[side]-finalVal[xside]) / 256;
#endif

#ifdef _DEBUG
    if( Alpha == -20000 ) {
        VDumpPawnMaps("!1", attacked_material, "attacked",
            &busy_pieces, "busy");
        printf("# val=%d ratioBonus=%d => %d (hkey=%I64x) centerpawn#=%d side=%d\n", delta, ratioBonus, 
            delta + ratioBonus, pste->hashkPc, phashp->c_center_pawns, side );
    	for (int co = coBLACK; co <= coWHITE; co++) {
            printf("# (%c) val=%d finalval=%d posval=%d tacval=%d endgval=%d other=%d \n",
                "BW"[co], val[co], finalVal[co], posval[co], tacval[co], endgval[co], otherval[co]);
            printf("# dynamics  attack=%d c_mob=%d mobility=%d attack#=%d\n", 
                attackval[co], 
                c_mobility[co], mobility[co], c_attack[co]);
            printf("# others : space in side=%d\n", defSpace[co]);
            printf("# pawns : val=%d  passed#=%d candidate#=%d backwards=%d connected=%d center=%d\n", phashp->side[co].val, 
                popCount(phashp->side[co].passed), popCount(phashp->side[co].candidate), 
                phashp->side[co].c_backward, phashp->side[co].c_connected, phashp->side[co].c_center );
            printf("# space Qside=%d center=%d Kside=%d\n",
                phashp->side[co].c_Qspace, phashp->side[co].c_Cspace, phashp->side[co].c_Kspace);
            printf("# Safety : attacks=%d defense=%d score=%d\n", c_kingAtk[co], c_kingDef[co], safetyval[co]);
#ifdef _DEBUG
            printf("#  mobility for pieces\n# ");
            for(int i = 0; i < mobDebugIdx; i ++)
                if( (mobDebugTable[i].piece & 1) == co ) 
                    printf("%c (%d)  ", stringPc[mobDebugTable[i].piece], mobDebugTable[i].mobility);
            printf("\n");
            fflush(stdout);
#endif
        }
    }
#endif
    stats.cFullEval++;
    pste->evalislazy = false;
    pste->evaluated  = true;
    if( (pste-1)->checkf && (delta < 0)) {
        int depth = int(pste - pcon->argste) + 1;
        if( depth >= 5 && (pste-3)->checkf && (pste-5)->checkf ) {
            int score = (delta + ratioBonus) / 2;
            pste->val = score;
            evalCache->set( pste, score);
            return score;
        }
    }
    int score;
    if( drawish )
        score = (delta + ratioBonus) / 2;
    else
        score = (delta ? delta : -2 ) + ratioBonus;
//    score = score & ~(2 - 1);
    evalCache->set( pste, score);
    pste->val = score;
    return score;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

/*
//  a   b   c   d   e   f   g   h
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	//
	//	White rook.
	//
//  a   b   c   d   e   f   g   h
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	//

*/

static U64 kposit[8];
static int pieceToPatBit[16];

static INLINE unsigned int squarePattern(const CON *pcon, int n, int f, int r, int sq) {
        kposit[n] = IdxToU64(sq);
        if(f < 0 || f > 7 || r < 0 || r > 7)
                return 1;
        else if(pcon->pieces[B_OCCU] & IdxToU64(sq))
                return 1;
        return 0;
}
static INLINE unsigned int squarePattern2(const CON *pcon, int n, int sq) {
        kposit[n] = IdxToU64(sq);
//      if(pieces[B_OCCU] & IdxToU64(sq))
//              return 1;
//      return 0;
        return pieceToPatBit[ pcon->pos[sq] ];
}
static INLINE unsigned int squarePattern3(const CON *pcon, int n, int sq) {
        kposit[n] = ((1LL)<<sq);
//      if(pieces[B_OCCU] & (C64(1)<<sq))
//              return 1;
//      return 0;
        // checking for white pieces is ok
//      if(pos[sq] & 1)
//              return 1;
        // checking for black pieces is not ok
        return pieceToPatBit[ pcon->pos[sq] ];
//      if(pos[sq] && !(pos[sq] & 1))
//              return 1;
//      return 0;
}

static unsigned int checkPattern(const CON *pcon) {
        int kpos = pcon->king[B];
        unsigned int pattern = 0;
        int f = File(kpos);
        int r = Rank(kpos);
//      pattern = (pattern << 1) | squarePattern(f,r,kpos);
        pattern = (pattern << 1) | squarePattern(pcon, 0, f-1,r,kpos-1);
        pattern = (pattern << 1) | squarePattern(pcon, 1, f-1,r+1,kpos+7);
        pattern = (pattern << 1) | squarePattern(pcon, 2, f,r+1,kpos+8);
        pattern = (pattern << 1) | squarePattern(pcon, 3, f+1,r+1,kpos+9);
        pattern = (pattern << 1) | squarePattern(pcon, 4, f+1,r,kpos+1);
        pattern = (pattern << 1) | squarePattern(pcon, 5, f-1,r-1,kpos-9);
        pattern = (pattern << 1) | squarePattern(pcon, 6, f,r-1,kpos-8);
        pattern = (pattern << 1) | squarePattern(pcon, 7, f+1,r-1,kpos-7);
        return pattern;
}

static unsigned int checkPattern2(const CON *pcon) {
        int kpos = pcon->king[B];
        unsigned int pattern = 0;
        int f = File(kpos);
        int r = Rank(kpos);
        if( f == 0 || f == 7 || r == 0 || r == 7 )
                return checkPattern(pcon);

        pattern = (pattern << 1) | squarePattern2(pcon, 0, kpos-1);
        pattern = (pattern << 1) | squarePattern2(pcon, 1, kpos+7);
        pattern = (pattern << 1) | squarePattern2(pcon, 2, kpos+8);
        pattern = (pattern << 1) | squarePattern2(pcon, 3, kpos+9);
        pattern = (pattern << 1) | squarePattern2(pcon, 4, kpos+1);
        pattern = (pattern << 1) | squarePattern2(pcon, 5, kpos-9);
        pattern = (pattern << 1) | squarePattern2(pcon, 6, kpos-8);
        pattern = (pattern << 1) | squarePattern2(pcon, 7, kpos-7);
        return pattern;
}

static unsigned int checkPattern3(const CON *pcon) {
        int kpos = pcon->king[B];
        unsigned int pattern = 0;
        int f = File(kpos);
        int r = Rank(kpos);

        if( kpos == G8 ) {
                pattern = (pattern << 1) | squarePattern3(pcon, 0, F8);
                pattern = (pattern << 1) | 1;
                pattern = (pattern << 1) | 1;
                pattern = (pattern << 1) | 1;
                pattern = (pattern << 1) | squarePattern3(pcon, 4, H8);
                pattern = (pattern << 1) | squarePattern3(pcon, 5, F7);
                pattern = (pattern << 1) | squarePattern3(pcon, 6, G7);
                pattern = (pattern << 1) | squarePattern3(pcon, 7, H7);
        }
        return pattern;

}

struct bitPat {
        unsigned char bishop_attack : 1;
        unsigned char queen_attack : 1;
        unsigned char pawn_attack : 1;
        unsigned char knight_attack : 1;
        unsigned char square_defended : 1;
} ;

typedef struct {
        struct bitPat   atkdef[8];
        unsigned char   score;
        unsigned char   atk_count;
        unsigned char   def_count;
} patEntry;

static patEntry        chunks[256];

static int checkKingAtk(const CON *pcon) {
        int idx = checkPattern3(pcon);
        patEntry        pat = chunks[idx];
        U64 bishop_attack = rand();
        U64 queen_attack = rand();
        U64 knight_attack = rand();
        U64 all_defense = rand();
        int atk_nok = 0;
        int def_nok = 0;
        for(int i = 0; i < 8 ; i++) {
                U64 sqMask = kposit[i];
                if( pat.atkdef[i].bishop_attack )
                        if( (bishop_attack & sqMask) == 0 )
                                atk_nok++;
                if( pat.atkdef[i].queen_attack )
                        if( (queen_attack & sqMask) == 0 )
                                atk_nok++;
                if( pat.atkdef[i].knight_attack )
                        if( (knight_attack & sqMask) == 0 )
                                atk_nok++;
                if( pat.atkdef[i].square_defended == 0 )
                        if( (all_defense & sqMask) != 0 )
                                def_nok++;
        }

        int atk_ok = (pat.atk_count - atk_nok);
        int def_ok = 8 - def_nok;
//      return (atk_ok / pat.atk_count) * (def_ok) / 8;
        return (32*atk_ok*def_ok) / pat.atk_count;
} 

