//
// Cyrano Chess engine
//
// Copyright (C) 2007,2008  Harald JOHNSEN
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
//
//

#include "engine.hpp"
#include "gproto.hpp"
#include "eval.hpp"
#include "moves.hpp"
#include "genmagic.hpp"

#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

#ifdef	DEBUG
static char const s_aszModule[] = __FILE__;
#endif

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-


static HASHPawn *s_rghashp;			// Pawn hash table.
static unsigned	s_chashpMac = 0;	// Number of elements in the table, minus one.  I
							//  do the minus one thing in order to avoid an
							//  inner-loop subtract, which is no big deal, but
							//  every little bit helps.
static bitboard centerControlMask[2];
static bitboard pawnon6or7Mask[2];
static bitboard pawnFileMask[8];
static bitboard pawnAdjacentFileMask[8];
static bitboard pawnAttackSpanMask[2][64];
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

#ifdef _DEBUG

//	I sat here and thought about how to write this routine, for about five
//	minutes.  I could have written something that would have required
//	modification any time the pawn structure stuff is messed with, but I
//	didn't like that idea, since the code I've used in the past for this
//	routine gets kind of annyong to change.

//	On the other hand, this program is supposed to be simple and direct.

//	Finally, I could not resist doing something I have never done -- write
//	something that works like "printf".

//	The format string is comprised of commands.  The commands are as follows:

//		digit ('0'..'9')		The next argument is an array of "digit"
//								bitmaps, to be displayed vertically, with
//								"*" for any set bit and "." for any zero bit.
//								The argument after this is a string pointing
//								to the column header.

//		'!'						The next argument is an array of two bitmaps,
//								which should be displayed as one bitmap with
//								"P" if a bit from the first map is set, and
//								"p" if a bit from the second map is set, and
//								"." if neither bit is set.  There is nothing
//								special if both maps are set, that comes up
//								"P".  The argument after this is a string
//								pointing to the column header.

//	The column header strings should be <= 8 characters long.

//	Here's an example call:

//		VDumpPawnMaps("!1", argbmLoc, "Loc", &bmDoubled, "Doubled");

//	This would be somewhat different, but would work:

//		VDumpPawnMaps("21", argbmLoc, "Loc", &bmDoubled, "Doubled");

//	There's an example line somewhere in the code below, unless I accidentally
//	deleted it.  You can comment it out and mess with it, if you decide you
//	want to add more pawn bitmaps and dump them.

//	I'm glad I wrote this routine, rather than just assuming that my code
//	worked.  It found a bunch of bugs.



void VDumpPawnMaps(char * szFmt, ...)
{
	va_list	pvaArg;
	char * sz;

	//	Display headers and figure out how many bitmaps in the largest array.
	//
	int cbmMax = 0;
	va_start(pvaArg, szFmt);
	for (sz = szFmt; *sz != '\0'; sz++) {
		char *	szArg;
		bitboard *pbmArg;
		int	cbm;

		if (isdigit(*sz))
			cbm = *sz - '0';
		else {
			Assert(*sz == '!');
			cbm = 1;
		}
		if (cbm > cbmMax)
			cbmMax = cbm;					// <-- New largest.
		pbmArg = va_arg(pvaArg, bitboard *);		// Pull bitmap pointer.
		szArg = va_arg(pvaArg, char *);		// Pull column header.
		if (sz != szFmt)
			putchar(' ');
		printf("%-8s", szArg);				// Display arg.
	}
	putchar('\n');
	//
	//	Display the bitmaps.  The largest argument I got was "cbmMax", so I
	//	will loop that many times, displaying one row of bitmaps each time.
	//
	for (int ibm = 0; ibm < cbmMax; ibm++) {

		if (ibm != 0)
			putchar('\n');
		//
		//	Display a row of bitmaps.
		//
		for (int rnk = rnk8; rnk >= rnk1; rnk--) {
			//
			//	Display one rank from all the bitmaps.
			//
			va_start(pvaArg, szFmt);
			for (sz = szFmt; *sz != '\0'; sz++) {
				char *	szArg;

				if (sz != szFmt)
					putchar(' ');
				int cbm = (isdigit(*sz)) ? *sz - '0' : 1;
				bitboard *pbmArg = va_arg(pvaArg, bitboard *);
				szArg = va_arg(pvaArg, char *);
				if (ibm >= cbm)			// This sometimes leaves trailing
					printf("        ");	//  spaces, but it's not worth fixing.
				else		// Display one rank from one bitmap.
					for (int fil = filA; fil <= filH; fil++) {
						bitboard bm;

						int isq = Idx(rnk, fil);
						bm = IdxToU64(isq);
						if (*sz != '!')
							putchar((bm & pbmArg[ibm]) ? '*' : '.');
						else if (bm & pbmArg[coWHITE])
							putchar('P');
						else
							putchar((bm & pbmArg[coBLACK]) ? 'p' : '.');
					}
			}
			putchar('\n');
		}
	}
}
#endif

static const int pawnDir[coMAX] = {16,-16};

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Penalties for pawn defects.

#define	valDOUBLED_ISO	20	// Doubled isolated pawn.
#define	valISOLATED		15	// Strictly isolated pawn (no pawns of this color
							//  on either adjacent file).
#define valISOLATED_OPEN 15 // isolated pawn on semi-open file (*added* to other penalities)

#define valBACKWARD     3
#define valBACKWARDweak 7

// Bonus for pawns

#define valCONNECTED    3   // this will be counted twice (once for each pawn)

// Normal doubled pawn, not strictly isolated. (idx=pawn count)
static const int doubledPenality[9] = {0,0,40,35,30,25,20,15,10};

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
static int tempTable[64];

// SetDistance/SetNextDistance is used to compute a path for a king from a square to another square
// The distance between a square and the destination square is stored in a square array
// so the distance can be known for each square of the board with a simple table lookup
//
void setNextDistance(int rank, int file, const CON *pcon, HASHPawn *phashp, int side, int distance) {
    // check if on board
    if( rank < 0 || rank > 7 || file < 0 || file > 7 )
        return;
    int from = Idx(rank, file);
    if( distance ) {
        // exit if square is unavailable (pawn or attacked by a pawn)
        if( (pcon->pos[from]|1) == (PAWN|1) )
            return;
        bitboard thisSqr = IdxToU64(from);
        if( phashp->side[side^1].pawnatk & thisSqr )
            return;
    }
    // update distance (depending on path taken)
    // exit if a shorter was leading to this square
    if( distance >= tempTable[from] )
        return;
    tempTable[from] = distance;
    if( distance < 15 )
        distance ++;
    setNextDistance(rank+1, file+0, pcon, phashp, side, distance);
    setNextDistance(rank-1, file+0, pcon, phashp, side, distance);
    setNextDistance(rank+0, file+1, pcon, phashp, side, distance);
    setNextDistance(rank+0, file-1, pcon, phashp, side, distance);
    setNextDistance(rank+1, file+1, pcon, phashp, side, distance);
    setNextDistance(rank+1, file-1, pcon, phashp, side, distance);
    setNextDistance(rank-1, file+1, pcon, phashp, side, distance);
    setNextDistance(rank-1, file-1, pcon, phashp, side, distance);
}
void setDistance(const CON *pcon, HASHPawn *phashp, int from, squareDist *keySq, int winningSide) {
    keySq->square = from;
    keySq->winningSide = winningSide;
    for(int side = 0; side <= 1; side ++) {
        for(int i = 0; i <= 63; i++)
            tempTable[i] = 15;
        setNextDistance(Rank(from), File(from), pcon, phashp, side, 0);
        memcpy(keySq->distance[side], tempTable, sizeof(keySq->distance[side]));
#ifdef DEBUG
        printf("\n");
	    for (int rnk = rnk8; rnk >= rnk1; rnk--) {
		    for (int fil = filA; fil <= filH; fil++) {
			    int isq = Idx(rnk, fil);
                if( tempTable[isq] == 15)
                    printf(" -");
                else if( tempTable[isq] )
			        printf(" %d", tempTable[isq]);
                else
                    printf(" o");
		    }
		    printf("\n");
	    }
#endif
    }
}

// from Steffan Westcott / Gerd Isenberg
// http://chessprogramming.wikispaces.com/All+Shortest+Paths
/////////////////////////////////////////////////////////////////////////
//
// Returns length of shortest path of set bits present in 'path' that
// 8-way connect any set bit in sq1 to any set bit of sq2. 0 is returned
// if no such path exists. Also fills a sequence of bitboards asp[length]
// that describes all such shortest paths.
 
int allShortestPaths(U64 sq1, U64 sq2, U64 path, U64* asp)
{
    // Do an 8-way flood fill with sq1, masking off bits not in path and
    // storing the fill frontier at every step. Stop when fill reaches
    // any set bit in sq2 or quit if fill cannot progress any further.
    // Then do 8-way flood fill from reached bits in sq2, ANDing the
    // frontiers with those from the first fill in reverse order.
 
    if (!(sq1 &= path) || !(sq2 &= path)) return 0;
                        // Drop bits not in path
                        // Early exit if sq1 or sq2 not on any path
    int i = 1;
    asp[0] = sq1;
 
    while(1)  // Fill from all set bits in sq1, to any set bit in sq2
    {
        if (sq1 & sq2) break;                     // Found good path
        U64 temp = sq1;
        sq1 |= eastOne(sq1) | westOne(sq1);       // Set all 8 neighbours
        sq1 |= soutOne(sq1) | nortOne(sq1);
        sq1 &= path;                              // Drop bits not in path
        if (sq1 == temp) return 0;                // Fill has stopped
        asp[i++] = sq1 & ~temp;                   // Store fill frontier
    }
 
    int length = i;                               // Remember path length
    asp[--i] = (sq2 &= sq1);                      // Drop unreached bits
 
    while(i)  // Fill from reached bits in sq2
    {
        U64 temp = sq2;
        sq2 |= eastOne(sq2) | westOne(sq2);       // Set all 8 neighbours
        sq2 |= soutOne(sq2) | nortOne(sq2);
        sq2 &= path;                              // Drop bits not in path
        asp[--i] &= sq2 & ~temp;                  // Intersect frontiers
    }
    return length;
}

// This function returns the length of the shortest path, not the path.
// This is for pawn endgames.
// The initial path should contain the whole board minus occupied squares, minus attacked squares
// also it is not called for pawns defended by other pawns (sq2 is not in the path in this case)
int pathLength(U64 sq1, U64 sq2, U64 path)
{
    // Do an 8-way flood fill with sq1, masking off bits not in path and
    // storing the fill frontier at every step. Stop when fill reaches
    // any set bit in sq2 or quit if fill cannot progress any further.
    // Then do 8-way flood fill from reached bits in sq2, ANDing the
    // frontiers with those from the first fill in reverse order.

    if (!(sq1 &= path) || !(sq2 &= path)) return 99;
                        // Drop bits not in path
                        // Early exit if sq1 or sq2 not on any path
    int i = 0; //1;

    while(1)  // Fill from all set bits in sq1, to any set bit in sq2
    {
        if (sq1 & sq2) break;                     // Found good path
        U64 temp = sq1;
        sq1 |= eastOne(sq1) | westOne(sq1);       // Set all 8 neighbours
        sq1 |= soutOne(sq1) | nortOne(sq1);
        sq1 &= path;                              // Drop bits not in path
        if (sq1 == temp) return 99;                // Fill has stopped
        i++;
    }
//    VDumpPawnMaps("1", &sq1, "path");
    int length = i;                               // Remember path length
    return length;
}

static int pawnSpace(const CON *pcon, int co, int file) {
    int space = 0;
    for(int rank = 1; rank <= 6; rank++) {
        int r = co == W ? rank : 7-rank;
        U8 c = pcon->pos[ Idx(r,file) ];
        if( (c|1) == (PAWN|1) )
            break;
        space++;
    }
    return space;
}

//	This gets the pawn structure info from the hash table if it's there,
//	otherwise it generates it.

//	In most cases, the stuff is in the hash table.  The hash table is big
//	enough, and the pawn structures don't change that much.

//	Since the pawn computation stuff isn't done that much, I don't take
//	special pains to make sure that it's fast.  It's better to be clear and
//	accurate.  It would be very possible to generate the defect values via
//	less complicated data structures, and fewer passes over the data, but
//	the excess time shoudn't matter.

//	On the other hand, the whole thing stays flexible so it should be easy
//	to add to.

//	This routine returns values from the perspective of the side to move,
//	and stores values in the hash table from white's perspective.

//	There are a lot of other pawn features that can be detected, but doubled
//	and isolated pawns should be good enough for now.  The trick is to play
//	solidly enough that the program doesn't drop pawns a few moves over its
//	horizon.

//	An obvious addition is *anything* that would help with the passed pawn
//	problem, which is always severe in a simple program.  The program could
//	also benefit from an understanding of backwardness and artificial
//	isolation.  Additionally, the values given for doubled and isolated pawns
//	could vary depending upon where the pawns are on the board, and the
//	condition of the opponent's pawns (a doubled isolated pawn on an open file
//	is worse than one on a closed file).

//	In a heavier implementation, some of the pawn maps might end up being
//	stored in the hash table itself, which is another level of complication.

static void ValPawns(const CON *pcon, const STE *pste)
{
    HASHPawn *phashp = &s_rghashp[pste->hashkPn & s_chashpMac];

    int	argval[coMAX];              // Total value of pawn features.
    bitboard argbmLoc[coMAX];       // Bits set for pawns that exist.
    bitboard bmDoubled = 0;                  // Bit set indicates doubled pawn.
    bitboard bmIsolated = 0;                 // Bit set indicates isolated pawn.
    bitboard bmInfluence[coMAX] = {0,0};
    bitboard    blocked[coMAX];

#ifdef KEYSQUARE
    phashp->keySquare[0].square = 0;
    phashp->keySquare[2].square = 0;
    phashp->keySquare[3].square = 0;
#endif
    argbmLoc[W] = pcon->pieces[WP];
    argbmLoc[B] = pcon->pieces[BP];
    phashp->side[W].location = argbmLoc[W];
    phashp->side[B].location = argbmLoc[B];
    phashp->pawnon6or7 = false;
    phashp->c_center_pawns = popCount( (argbmLoc[W] | argbmLoc[B]) & centerMask);
    // new version with more setwise operations, thx Gerd
    blocked[W] = argbmLoc[W] & bFrontSpans(argbmLoc[B]);
    blocked[B] = argbmLoc[B] & wFrontSpans(argbmLoc[W]);

    for (int coUs = coBLACK; coUs <= coWHITE; coUs++) {
        phashp->side[coUs].c_doubled = 0;
        phashp->side[coUs].c_islands = 0;
        phashp->side[coUs].pawnatk   = 0;
        phashp->side[coUs].candidate = 0;
        phashp->side[coUs].c_backward= 0;
        phashp->side[coUs].c_connected=0;
        phashp->side[coUs].c_center  = 0;
        phashp->side[coUs].c_Qspace  = 0;
        phashp->side[coUs].c_Cspace  = 0;
        phashp->side[coUs].c_Kspace  = 0;
        phashp->pawnon6or7 = phashp->pawnon6or7 || (pawnon6or7Mask[coUs] & argbmLoc[coUs]);
        // count the number of pawns on this file
        for(int f = 0; f < 8 ; f++)
            phashp->side[coUs].c_pawn[f] = 0;

        argval[coUs] = 0;
        // build the influence bitmap, squares that are attacked by pawns
        // on their position or if they move up/down
        // note that pawns can move from one file to another so this computation
        // can be wrong in some situations
        int dir = coUs == W ? 8 : -8;

        bitboard xpawns = argbmLoc[coUs^1];
        for (bitboard pawns = argbmLoc[coUs]; pawns; ) {
            int sqr = PopLSB(&pawns);
            phashp->side[coUs].pawnatk |= pawn_attacks[coUs][sqr];
            int guard = 0;
            bmInfluence[coUs] |= pawnAttackSpanMask[coUs][sqr];
            bitboard tmp = pawnAttackSpanMask[coUs][sqr] & xpawns;
            if( tmp ) {
                guard ++;
                tmp = tmp & (tmp - 1);
                if( tmp )
                    guard++;
            }
            if( guard <= 1 && ((blocked[coUs] & IdxToU64(sqr)) == 0))
                phashp->side[coUs].candidate |= IdxToU64(sqr);
            phashp->side[coUs].c_pawn[File(sqr)]++;
        }

        //	Collect doubled pawn info.  I'm see if there is more than one
        //	pawn of a certain color on the a-file, if there is I will OR
        //	all of those pawns into the "bmDoubled" bitmap, then I'll try
        //	the b-file, and so on.
        //
        // for each file...
        for (int f = 0; f <= 7; f++) {
            //	Get all pawns of this color on this file.
            //
            bitboard bmS = argbmLoc[coUs] & pawnFileMask[f];
            //
            //	If more than one bit set in this map, all the pawns of
            //	this color on this file are doubled.
            //
            if (bmS & (bmS - 1))
                bmDoubled |= bmS;
            //
            if (!(pawnAdjacentFileMask[f] & argbmLoc[coUs]))
                bmIsolated |= bmS;
        }
        // count space for each side and for each flank
        // space is the number of squares behind our pawns, or the number of squares
        // in front of the opponent pawns if we don't have a pawn on the file
        // So an open or semi-open file gives a lot of space and advancing pawns
        // gives some space too
        phashp->side[coUs].c_Qspace  = pawnSpace(pcon, coUs, filA)
                                    + pawnSpace(pcon, coUs, filB)
                                    + pawnSpace(pcon, coUs, filC);
        phashp->side[coUs].c_Cspace  = pawnSpace(pcon, coUs, filD)
                                    + pawnSpace(pcon, coUs, filE);
        phashp->side[coUs].c_Kspace  = pawnSpace(pcon, coUs, filF)
                                    + pawnSpace(pcon, coUs, filG)
                                    + pawnSpace(pcon, coUs, filH);
        phashp->side[coUs].c_space = (phashp->side[coUs].c_Qspace + 
            2*phashp->side[coUs].c_Cspace + 
            phashp->side[coUs].c_Kspace);
	}
    phashp->side[W].influence = bmInfluence[W];
    phashp->side[B].influence = bmInfluence[B];
    // queen side pawn majority
    int c_wq = lowpopCount( argbmLoc[W] & QSideMask );
    int c_bq = lowpopCount( argbmLoc[B] & QSideMask );
    if( c_wq > c_bq && ((argbmLoc[W] & QSideMask & bmDoubled) == 0 ) )
        argval[W] += 7;
    if( c_wq < c_bq && ((argbmLoc[B] & QSideMask & bmDoubled) == 0 ) )
        argval[B] += 7;

	for (int coUs = coBLACK; coUs <= coWHITE; coUs++) {
        phashp->side[coUs].passed = argbmLoc[coUs] & 
            (~blocked[coUs]) & (~phashp->side[coUs ^ 1].influence) ;
        // control of the center
        phashp->side[coUs].c_center = popCount( phashp->side[coUs].pawnatk & centerControlMask[coUs] );

        // for each pawn of the board...
        int valDoubled = 3 * doubledPenality[ pcon->msigCount[PAWN|coUs] ] / 4;
        for (bitboard pawns = argbmLoc[coUs]; pawns; ) {
            int sqr = PopLSB(&pawns);
            int nSqr = remapSq[coUs][sqr];
            bitboard bmLoc;
            bmLoc = IdxToU64(sqr);
            int f = File(sqr);
            if (bmLoc & bmIsolated) {
                if( phashp->side[coUs^1].c_pawn[f] == 0 )
                    argval[coUs] -= valISOLATED_OPEN;
                if (bmLoc & bmDoubled)
                    argval[coUs] -= valDoubled;
                argval[coUs] -= valISOLATED;
            } else if (bmLoc & bmDoubled) {
                argval[coUs] -= valDoubled;
            } else {
                int dir = coUs == W ? -1 : 1;
                int r = Rank(sqr);
                bool hasFriends = false;
                while(r > 0 && r < 7 ) {
                    if( f > 0 && pcon->pos[Isq64FromRnkFil(r,f - 1)] == (PAWN|coUs)) {
                        hasFriends = true;
                        break;
                    }
                    if( f < 7 && pcon->pos[Isq64FromRnkFil(r,f + 1)] == (PAWN|coUs)) {
                        hasFriends = true;
                        break;
                    }
                    r += dir;
                }
                // connected pawns
                if( hasFriends ) {
                    if( r == Rank(sqr) ) {
                        phashp->side[coUs].c_connected++;
                        argval[coUs] += valCONNECTED;
                    }
                // non isolated pawn on an open file
                } else if( phashp->side[coUs^1].c_pawn[f] == 0 ) {
                    // if there is no friendly pawn behind this one then it's backward
                    // c_backward is used for *debug only*
//                    phashp->side[coUs].c_backward ++;
                    argval[coUs] -= valBACKWARD;
                    dir = -dir;
                    r = Rank(sqr) + dir;
                    while(r > 0 && r < 7 ) {
                        // check for ennemy pawn
                        if( f > 0 && pcon->pos[Isq64FromRnkFil(r + dir,f - 1)] == (PAWN|coUs^1)) {
                            argval[coUs] -= valBACKWARDweak;
//                            phashp->side[coUs].c_backward += 10;
                            break;
                        }
                        if( f < 7 && pcon->pos[Isq64FromRnkFil(r + dir,f + 1)] == (PAWN|coUs^1)) {
                            argval[coUs] -= valBACKWARDweak;
//                            phashp->side[coUs].c_backward += 10;
                            break;
                        }
                        if( f > 0 && pcon->pos[Isq64FromRnkFil(r,f - 1)] == (PAWN|coUs))
                            break;
                        if( f < 7 && pcon->pos[Isq64FromRnkFil(r,f + 1)] == (PAWN|coUs))
                            break;
                        r += dir;
                    }
                }
            }
		}
	}
#ifdef KEYSQUARE
//    setDistance(pcon, phashp, G5, &phashp->keySquare[0], W);
//    setDistance(pcon, phashp, B5, &phashp->keySquare[1], W);
    setDistance(pcon, phashp, A5, &phashp->keySquare[0], W);
    setDistance(pcon, phashp, F5, &phashp->keySquare[1], W);
#endif

#ifdef _DEBUG
    static int dump_count = 0;
    dump_count++;
    if( dump_count <= 1 ) {

//        bitboard blocked[2] = {phashp->side[0].blocked,phashp->side[1].blocked};
        bitboard passed[2] = {phashp->side[0].passed, phashp->side[1].passed };
        VDumpPawnMaps("!111111",
            argbmLoc, "Loc",
            &bmDoubled, "Doubled",
            &bmIsolated, "Isolated",
            &phashp->side[W].influence, "WInfl.",
            &phashp->side[B].influence, "BInfl.",
//            &blocked, "blocked",
            &phashp->side[W].pawnatk, "WAtk.",
            &phashp->side[B].pawnatk, "BAtk."
        );
    }
#endif
    phashp->hashk = pste->hashkPn;
    // keep a score for each side
    phashp->side[W].val = argval[coWHITE];
    phashp->side[B].val = argval[coBLACK];

    // now we know if a pawn is
    // - isolated, doubled, blocked/candidate, passed
    // we also have the count of pawn per file
    // we can do rook_file_attack & location(xside) to know if a rook is on a semi open file
}

void EvalPawns(const CON *pcon, const STE *pste, PHASHPawn & phashp) {
	phashp = &s_rghashp[pste->hashkPn & s_chashpMac];

	if (phashp->hashk != pste->hashkPn) {
        ValPawns(pcon, pste);
    }
}


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Zero all pawn hash memory.

void VClearHashp(void)
{
    if( s_chashpMac )
	    memset(s_rghashp, 0, (s_chashpMac + 1) * sizeof(HASHPawn));
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This allocates the pawn hash memory according to the max value passed in
//	by the user, then clears the table.

bool FInitHashp(PCON pcon, int cbMaxPawnHash)
{
    centerControlMask[W] = centerControlMask[B] = 0;
    for(int f = 2; f <= 5; f++)
        for(int r = 3; r <= 6; r++) {
            centerControlMask[W] |= IdxToU64( Idx(r,f) );
            centerControlMask[B] |= IdxToU64( Idx(7-r,f) );
        }
    pawnon6or7Mask[B] = pawnon6or7Mask[W] = 0;
    for(int r = 0; r <= 7; r++) {
        pawnon6or7Mask[B] |= IdxToU64( A3+r );
        pawnon6or7Mask[B] |= IdxToU64( A2+r );
        pawnon6or7Mask[W] |= IdxToU64( A6+r );
        pawnon6or7Mask[W] |= IdxToU64( A7+r );
    }
    bitboard bmI = (U64)0x0101010101010101;	// Start with a1...a8.
    for(int f = 0; f <= 7; f++) {
        pawnFileMask[f] = bmI;
        pawnAdjacentFileMask[f] = 0;
        if (!(bmI & 0x01))
            pawnAdjacentFileMask[f] |= bmI >> 1;
        if (!(bmI & 0x80))
            pawnAdjacentFileMask[f] |= bmI << 1;
        bmI <<= 1; // next file
    }
    for(int co = 0; co <= 1; co++)
        for(int sqr = 0; sqr <= 63; sqr++) {
            if(co == W )
                pawnAttackSpanMask[co][sqr] = nortFill(pawn_attacks[co][sqr]);
            else
                pawnAttackSpanMask[co][sqr] = soutFill(pawn_attacks[co][sqr]);
        }

    if( cbMaxPawnHash <= 0 || cbMaxPawnHash > 256*1024*1024 )
        cbMaxPawnHash = 1*1024*1024;

    int chashpMax = 1;
	for (;;) {
		if (chashpMax * 2 * (int)sizeof(HASHPawn) > cbMaxPawnHash)
			break;
		chashpMax *= 2;
	}
    int Hsize = chashpMax * sizeof(HASHPawn);
	VPrSendComment("%d Kbytes pawn hash memory (%d entries of %d bytes)", Hsize/1024, chashpMax, sizeof(HASHPawn));
    s_rghashp = (HASHPawn *) malloc(Hsize);
	if ( s_rghashp == NULL) {
		VPrSendComment("Can't allocate pawn hash memory: %d bytes", Hsize);
        chashpMax = 1024;
        Hsize = chashpMax * sizeof(HASHPawn);
	    VPrSendComment("%d Kbytes pawn hash memory", Hsize/1024);
        s_rghashp = (HASHPawn *) malloc(Hsize);
	    if ( s_rghashp == NULL) {
		    VPrSendComment("Can't allocate pawn hash memory: %d bytes", Hsize);
		    return false;
        }
	}
	s_chashpMac = chashpMax - 1;
	VClearHashp();
	return true;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
