//
// Cyrano Chess engine
//
// Copyright (C) 2007  Harald JOHNSEN
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
//
//

#include "magicmoves.h"
#include "engine.hpp"
#include "gproto.hpp"
#include "hash.hpp"
#include "board.hpp"
#include "moves.hpp"
#include "stats.hpp"
#include "search.hpp"
#include "eval.hpp"
#include "genmagic.hpp"
#include <stdlib.h>

#ifdef	_DEBUG
static char const s_aszModule[] = __FILE__;
#endif

// enables the PVS search, undefine to have standard alpha beta search (for bug hunting)
// reduce the number of nodes
#define NULL_WINDOW

// reduce the number of nodes
#define NULL_MOVE

// allways extend depth by one ply when in check
// increase the number of nodes, find mates in less plies
#define EXTEND_CHECK
// extend the pv a bit, the score should be more reliable, and this gives a better move ordering for the next iteration
#define SOME_PV_EXTENSION
// reduce the number of nodes
#define USE_HASH
// change move ordering => reduce the number of nodes
#define USE_KILLERS

#define QS_BEFORE_HASH      // note:probing hash first gives worse results

// while it appears to work, fhr can not be implemented so simply
// I disable it for now
//#define FHR_THREAT  (300)
//#define FHR

// late move reduction : ok it's not so bad even with my poor implementation
#define USE_LMR // disabled to test things
#define USE_SIMPLE_LMR
#define REDUCE_PAWN // this gives a very nice boost
// skip descending the tree for some useless xchange in the quiescent search
// reduce the number of nodes ?
#define PRUNE_QS
// ...
#define FUTILITY_PRUNING
// this will stop search when we have a score good enought (mate)
#define DISTANCE_PRUNING

//#define PRUNE_SIBLING

#define RECAPTURE_EXTENSION
// change move ordering
#define HISTORY

//#define MOVE_SAFETY_SORT  // this is not so bad
//#define MOVE_PST_SORT     // this is not so good
//#define MOVE_FILTER       // not implemented, first tests were not good (don't know why)

#define USE_IID
//#define HASH_THREAT
//#define HASH_NM
#define EVAL_INTERIOR
#define USE_RECOGNIZER

#define moveisCapture(pcm) (Capt((pcm)->m) != 0)
#define moveisGoodCapture(pcm) (((pcm)->cmk & (cmkGOODCAPT|cmkEQUCAPT)) != 0)
#define moveisVeryGoodCapture(pcm) (((pcm)->cmk & (cmkGOODCAPT)) != 0)
#define moveisBadCapture(pcm) (((pcm)->cmk & (cmkBADCAPT)) != 0)
#define moveisPromotion(pcm) ((Pc((pcm)->m)|1) == (B_PROM|1))
#define moveisPawnMove(pcm) ((Pc((pcm)->m)|1) == (PAWN|1))
#define moveto7thRank(pcm)  ((Rank(To((pcm)->m)) == 6) || (Rank(To((pcm)->m)) == 1))
#define moveisGood(pcm) (((pcm)->cmk & (cmkMASK)) >= cmkQUEEN)


#define FRONTIER    0
#define isFrontier(pste)    ((pste)->plyRem == FRONTIER)
#define isRazorDepth(pste)    ((pste)->plyRem == (FRONTIER+2))

static int SearchQ(PCON pcon, PSTE pste, int Alpha, int Beta);
static int Search(PCON pcon, PSTE pste, int Alpha, int Beta, nodeType nt);
static int SearchPV(PCON pcon, PSTE pste, int Alpha, int Beta);
static bool inline capture_is_good(PCON pcon, const STE *pste, const CM *pcm);


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

static void SortKillers(PSTE pste, int depth)
{
    U32 Killer1 = Killers[depth].Killer1.m;
    U32 Killer2 = Killers[depth].Killer2.m;
    U32 Killer3 = Killers[depth-2].Killer1.m;
    U32 Killer4 = Killers[depth-2].Killer2.m;
    foreachmove(pste, pcm)
        if( (pcm->m == MateKillers[depth].Killer1.m) || (pcm->m == MateKillers[depth].Killer2.m)) {
			pcm->cmk |= cmkKILLER | 5;
            stats.Killers++;
        } else if ( (pcm->m == Killer1) || (pcm->m == Killer2) ) {
			pcm->cmk |= cmkKILLER | 3;
            stats.Killers++;
		} else if ( (pcm->m == Killer3) || (pcm->m == Killer4) ) {
			pcm->cmk |= cmkKILLER | 1;
            stats.Killers++;
		}
}

static void SortKingAttack(PSTE pste)
{
    foreachmove(pste, pcm) {
    }
}

static inline int MaterialGain(const CON *pcon, const CM *pcm) {
    // Capt(move) == 0 => no capture, array return 0 in this case
    return b_valPc[ Capt(pcm->m) ];
}


//	This will be used to figure out which move to search first.  It is very
//	important to try to search the best move first, since this will produce
//	a huge reduction in tree size.

//	Best moves are typically winning captures and moves that have been shown
//	to be good here in the past (PV move and hash table move).

static void PrepareSort(CON *pcon, const STE *pste, int depth, PHASHPawn phashp) {
    // bitmap of squares attacked by the opponent pawns
    bitboard notSafe = phashp->side[pste->side ^1 ].pawnatk;
//    U32 lm = 0;
//    if(depth >= 2)
//        lm = (pste-1)->lastMove;

    foreachmove(pste, pcm) {
        pcm->cmk = 0;
        if(Capt(pcm->m)) {
            pcm->cmk = MVV_LVA[Pc(pcm->m)][Capt(pcm->m)];
            if( pcm->cmk & cmkBADCAPT ) {
                // RxN, QxP
                if( capture_is_good(pcon, pste, pcm ) ) {
                    pcm->cmk = cmkGOODCAPT | 2000;
                }
            } 
        }
        else if( moveisPawnMove(pcm) && moveto7thRank(pcm)) {
            // only checking the dest rank because we know it's a pawn
//            pcm->cmk = (cmkKILLER | cmkGOODCAPT) + 10000;
            pcm->cmk = cmkQUEEN;
        } else if( moveisPromotion(pcm) ) {
            if( Prom(pcm->m) == PROM_Q )
                pcm->cmk = cmkGOODCAPT | cmkQUEEN;
            else
                pcm->cmk = cmkQUEEN | (100 - Prom(pcm->m));
        } else {
            // having 'bad' captures after non capture moves gives more nodes
            // ie having non captures after 'bad' captures
            pcm->cmk = cmkNONE;
            int history_score = 0;
#ifdef HISTORY
//            pcm->cmk |= History.sort_key( pcm );
            history_score = History.get( pcm, depth );
            pcm->cmk |= 1000 + history_score;
#endif

#ifdef MOVE_SAFETY_SORT
            if( history_score == 0 )
            {
                if(!moveisPawnMove(pcm)) {
                    bitboard to = IdxToU64(To(pcm->m));
                    // keep key positive
                    pcm->cmk += 1;
                    // moving to a square attacked by a pawn
                    if( notSafe & to )
                        pcm->cmk -= 1;
                    bitboard from = IdxToU64(From(pcm->m));
                    // leaving a square attacked by a pawn
                    if( notSafe & from )
                        pcm->cmk += 1;
                }
    //        if( Attacked(pcon, To(pcm->m), pste->side ^ 1 ) )
    //            pcm->cmk -= 10;

    #ifdef MOVE_FILTER
            // not implemented, ideas are
            //  malus is moving the same piece twice in a row, moving back to the last position
            //  bonus if the move can be done because another piece moved from To(m) or was
            //  between From(m) & To(m)
            if(lm) {
                if( Pc(pcm->m) == Pc(lm) ) {
                    //pcm->cmk -= 4;
                    if( To(pcm->m) == From(lm) )
                        pcm->cmk -= 8;
                } else {
                }
            }
    #endif

        }
#endif
#ifdef MOVE_PST_SORT
        if( history_score == 0 ) {
            // set cmk to delta positional value
            int pc = Pc(pcm->m);
            int neutralPc = pc & (~1);
            int co = pc & 1;
            // neutral color square for positional values
            int nSqr = remapSq[co][To(pcm->m)];
            // basic positional value
            pcm->cmk += staticPosVal[neutralPc >> 1][nSqr];
            nSqr = remapSq[co][From(pcm->m)];
            // basic positional value
            pcm->cmk -= staticPosVal[neutralPc >> 1][nSqr];
        }
#endif
        }

    }
}

static void PrepareSortQS(PCON pcon, const STE *pste, int depth) {
    foreachmove(pste, pcm) {
        U32 m = pcm->m;
        pcm->cmk = 0;
        if(Capt(m)) {
            pcm->cmk = MVV_LVA[Pc(m)][Capt(m)];
            if( pcm->cmk & cmkBADCAPT ) {
                // RxN, QxP
                if( capture_is_good(pcon, pste, pcm ) ) {
                    pcm->cmk = cmkGOODCAPT | 2000;
                }
            }
        }
    }
}

static inline void SortNext(const STE *pste, PCM pcm) {
	PCM	pcmBest = pcm;

	for (PCM pcmI = pcm + 1; pcmI < (pste + 1)->mlist_start ; pcmI++)
		if (pcmI->cmk > pcmBest->cmk)
			pcmBest = pcmI;
	CM cm = *pcm;
	*pcm = *pcmBest;
	*pcmBest = cm;
}

static inline int MaterialValue(const STE *pste) {
    return (pste->valPcUs + pste->valPnUs) - (pste->valPcThem + pste->valPnThem);
}

static inline int recapMaterialValue(const CON *pcon, const CM *pcm) {
    // Capt(move) == 0 => no capture, array return 0 in this case
    return exchangePieceValue[ Capt(pcm->m) ];
}

static inline void copyPV(PSTE pste, const CM &cm) {
    int how_many = (pste + 1)->ccmPv;
    if( how_many > 0 )
        memcpy(pste->argcmPv + 1, (pste + 1)->argcmPv, how_many * sizeof(CM));
    pste->argcmPv[0] = cm;
    pste->ccmPv = how_many + 1;
}
static inline void copyPVQS(PSTE pste, const CM &cm) {
#if 0
    int how_many = (pste + 1)->ccmPv;
    if( how_many > 0 )
        memcpy(pste->argcmPv + 1, (pste + 1)->argcmPv, how_many * sizeof(CM));
    pste->argcmPv[0] = cm;
//    pste->argcmPv[0].cmk = cmkHASH | 123456;
    pste->ccmPv = how_many + 1;
#else
    pste->argcmPv[0] = cm;
#endif
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This does a null-move search to see if the side not to move has a threat.
//	If they can't drive the score above alpha, even if they get to move twice
//	in a row, their positions probably sucks, and we'll prune this variation.
//	It's not called in endgames, because it has problems detecting zugzwang.

//  We don't need to do a null move if we know allready that there is threats around
//  In endgames any pawn move is a potential threat (score will vary a lot)

static int FThreat(PCON pcon, PSTE pste, int Alpha, int Beta, const nodeType nt)
{
	bNullMove(pcon, pste);
    // Reduced-depth search
    int R = 2;
    // we can reduce the depth further when there is lot of material on the board
    // this does not work, I get a lot more nodes
    // R = 3 if (rem_ply > 8) || (rem_ply > 6 (and pieces per side >= 3))
#if 1
    if( (pste->plyRem >= 5) || (pste->valPcThem + pste->valPcUs >= 6*valKNIGHT) 
                                && (pste->plyRem >= 4) ) {
        R = 3;
    }
#endif
    R = 3; // set to 3 since we do a verification search
    (pste + 1)->plyRem = pste->plyRem - 1 - R;
#if 0
    if(pste->plyRem >= 4) {
        int val = -SearchQ(pcon, pste + 1, - pste->val, - Beta + 1); /* QS-level search of reply */
        if( val >= Beta + (3*valPAWN)*(pste->plyRem-3)/2 )
            return val;    /* hyper NMP (= verified razoring?) */
    } 
#endif
    int val = -Search(pcon, pste + 1, - Beta, - Beta + 1, nt);
	return val;
}


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Quiescent search.  This lets captures play out at the end of the search,
//	so I don't end up aborting a line at a stupid place.

//	The quiescent search doesn't see checks, so it will not detect checkmates,
//	and might actually end up making decisions based upon an illegal line.

//	This is one reason I don't record the PV while in quiescent search -- I
//	don't want people telling me the program is doing illegal stuff.

// 200 gives better results than nothing or 100 or 150 (on wac-141)
//#define QS_MARGIN   200//100
//#define QS_MARGIN   100
//#define QS_MARGIN   125
//#define QS_MARGIN   200
#define QS_MARGIN   75

static bool inline capture_is_good(PCON pcon, const STE *pste, const CM *pcm) {
    int pc = Pc(pcm->m);
    int took = Capt(pcm->m);
    if( (pc|1) == (KING|1) )
        return true;
    if( moveisPromotion( pcm ) )
        return true;
    if( exchangePieceValue[ took ] >= exchangePieceValue[ pc ] )
        return true;
    // en passant (note that this is allready handled by the line above)
    // because took == 0 == black king value
    if( (pc|1) == (B_EP|1) )
        return true;

    // compute see value...
    int val = see(pcon, pste, From(pcm->m), To(pcm->m));
    return val >= 0;
}

static char sz_score[128];

static int SearchQ(PCON pcon, PSTE pste, int Alpha, int Beta)
{
	pcon->ss.nodes++;
    stats.QS_count ++;
	CM cmBest(0,0);
	pste->ccmPv = 0;
	bool fFound = false;
    bool reply_to_check = false;
    bool isPV = Beta != (Alpha+1);
    int iaminCheck = pste->checkf;
    int  f_score = 0;
    int  f_max   = valMAX;
    CM * p_moves;
    int depth = int(pste - pcon->argste) + 1;
#ifdef DISTANCE_PRUNING
#if 1
    // this is the lowest score I can get from here
    int i_am_mate_score = mateValue(depth + 1);
    if( i_am_mate_score > Alpha ) {
        Alpha = i_am_mate_score;
        // Beta is mate in n, best here is mate in n+m, so return
        if(i_am_mate_score >= Beta)
            return Beta; //i_am_mate_score;
//            return i_am_mate_score;
    }
#endif
#endif
	//
	//	See if static eval will cause a cutoff or raise alpha.
	//
    // this will be the return value if king evasion does no capture
	int posVal = ValEval(pcon, pste, Alpha, Beta);
    // score is valMIN when in check or else it won't play moves that lead to loss of material when nearly mated
    int bestVal = valMIN; //posVal;


#ifdef PRUNE_QS
    if( !iaminCheck ) {
        f_score = posVal + QS_MARGIN ;
        f_max   = f_score;
    }
#endif
    if( !iaminCheck ) {
        bestVal = posVal;
	    if (bestVal >= Beta)
		    return bestVal;
        if (bestVal > Alpha)
            Alpha = bestVal;
    }
    // generate captures, if we are in check then generate all moves (evade)
	if (iaminCheck) {
		bGenEvasions(pcon, pste);
#if 1
        if( legalMoveCount( pste ) == 1 )
            reply_to_check = true;
        else
            bestVal = posVal;
#endif
        if( legalMoveCount( pste ) == 0 && (pste->plyRem != -1))
            return posVal;
	}
	else {
		p_moves = bGenCaptures(pcon, pste, pste->mlist_start);
	}

    if( depth > pcon->ss.plyMaxDepth )
        pcon->ss.plyMaxDepth = depth;
    PrepareSortQS( pcon, pste, depth );
    //
	//	Iterate through capturing moves, to see if I can improve upon alpha
	//	(which may be the static eval).
	//

    foreachmove(pste, pcm) {

        SortNext(pste, pcm);
        fFound = true;
        bMove(pcon, pste, pcm->m);
#if 1
        // we've generated all moves if in check, but we wont play them
        // we just look if we have a legal move
        if( iaminCheck && !reply_to_check && (Capt(pcm->m) == 0) ) {
            (pste+1)->ccmPv = 0;
		    bUndoMove(pcon, pste, pcm->m);
            continue;
        }
#endif
#ifdef PRUNE_QS
        if( !iaminCheck 
//            && (pste[1].checkf == 0 || (pste->plyRem != -1) )
//            && (!isPV)
            && !moveisPromotion(pcm) && !moveisPawnMove(pcm)) {
                if( moveisBadCapture(pcm) 
                    /*&& (pste[1].checkf == 0 || (pste->plyRem != -1) )*/) {
                    stats.FP_QS++;
    		        bUndoMove(pcon, pste, pcm->m);
                    break;
                }
                if (f_max + MaterialGain(pcon, pcm) <= Alpha) {
                    stats.FP_QS++;
    		        bUndoMove(pcon, pste, pcm->m);
                    // check that
                    int approx = f_max + MaterialGain(pcon, pcm);
                    if( approx > bestVal ) {
                        cmBest.m = 0;
	                    pste->ccmPv = 0;
                        // is that needed ?
                        bestVal = approx;
                    }
                    continue;
                }
        }
#endif

		(pste + 1)->plyRem = pste->plyRem - 1;
		int val = -SearchQ(pcon, pste + 1, -Beta, -Alpha);
		bUndoMove(pcon, pste, pcm->m);
#if 0
        if( pcon->ss.plyDepth == 1 ) {
            sprintf(sz_score, " QS v=%d [%d,%d] ", val, Alpha, Beta);
            VDumpLine(pcon, pste, sz_score);
        }
#endif
        if (val >= Beta) {
            copyPVQS(pste, *pcm );
			return val;
        }
        if (val > bestVal) {
			bestVal = val;
            cmBest = *pcm;
            copyPVQS(pste, *pcm );
            if (val > Alpha)
			    Alpha = val;
        }
	}
#if 0
    if( !iaminCheck && (pste->plyRem == -1) && (bestVal < (valKNIGHT/2))
        && (pste->valPcUs >= valROOK)) {
        // gen checks
        // this is of course not optimal
        bGenNonCaptures(pcon, pste, p_moves);

        foreachmove(pste, pcm) {

            fFound = true;
            bMove(pcon, pste, pcm->m);
            // we only want moves that do check
            if( (pste[1].checkf == 0) ) {
		        bUndoMove(pcon, pste, pcm->m);
                continue;
            }
		    (pste + 1)->plyRem = pste->plyRem - 1;
		    int val = -SearchQ(pcon, pste + 1, -Beta, -Alpha);
		    bUndoMove(pcon, pste, pcm->m);
            if (val >= Beta) {
                copyPVQS(pste, *pcm );
			    return val;
            }
            if (val > bestVal) {
			    bestVal = val;
                cmBest = *pcm;
                copyPVQS(pste, *pcm );
                if (val > Alpha)
			        Alpha = val;
            }
	    }
    }
#endif

	if (!fFound) {
        // We have no move !!
		if (iaminCheck) {
            // Checkmate
			bestVal = mateValue(depth);
		}
            // perhaps draw but we did not gen all moves
	}
	return bestVal;
}


//	Full-width search.  This is the heart of the "chess playing" part of this
//	program.  It performs a recursive alpha-beta search.

#define LOWERBOUND  Alpha

#define NCMove(from, to, pc) ((from) | ((to) << 6) | ((pc) << 12))
#define CMove(from, to, type, capt) ((from) | ((to) << 6) | ((type) << 12) | ((capt) << 16))

static int Search(PCON pcon, PSTE pste, int Alpha, int Beta, nodeType nt)
{
	int	val;
    pste->ccmPv = 0;
    pste->reduced = false;
	CM cmBest(0,0);
    bool king_attak = false;
    int king_attacker = -1;
    int preExtend = 0;
    int bestVal = valMIN;
    int iaminCheck = pste->checkf;
    bool nm_threat = false;

    if (pste->plyRem < 0) {
        return SearchQ(pcon, pste,  Alpha, Beta);
    }

    int depth = int(pste - pcon->argste) + 1;
    if( depth >= (csteMAX-50) ) {
        return SearchQ(pcon, pste,  Alpha, Beta);
    }

	//	Increment global node counter, and figure out if it is time to call
	//	the interface callback, which is what allows for an interruptible
	//	search.  It's crucial that this be called now and then.
	//
	if (++pcon->ss.nodes >= pcon->ss.nodesNext) {
		//
		//	2000 nodes is maybe 1/50to 1/100 of a second.
		//
		pcon->ss.nodesNext = pcon->ss.nodes + 20000;
		VPrCallback(pcon);
		VCheckTime(pcon);
		if ((pcon->fAbort) || (pcon->fTimeout))
			return 0;
        if( pcon->smode == smodeTHINK || pcon->smode == smodeANALYZE ) 
            VPrSendUciCurrentLine(pcon, pste);
	}
	//
	//	I can return draws and cut off if I'm somewhere other than the
	//	first ply of depth.
	//
	//	Return draw score if this is a repeated node.
	//
	if (FRepSet(pcon, pste->hashkPc)) {
        stats.repCount++;
		return 0;
	}

    bool no_threat = true;
#ifdef EVAL_INTERIOR
    // gathering stats
    int posVal = pste->val;
    if(!pste->evaluated )
	    posVal = ValEval(pcon, pste, Alpha, Beta);
//    int nmPosVal = posVal + valPAWN / 4;
//    int nmPosVal = posVal - 15;
    int nmPosVal = posVal;
    if( false && !pste->evalislazy ) {
        // if we have some hanging material
        if( pste->threat[pste->side] ) {
            no_threat = false;
            stats.c_threat++;
            if( (pste-1)->reduced ) {
                // there will be a re-search
                return valMIN;
            }
        }
    }
    if( !pste->evalislazy && !(pste-1)->evalislazy && (Alpha < valKNIGHT)) {
#if 0
        if( (pste-1)->reduced && (pste->kingsafety[pste->side]+35) < (pste-1)->kingsafety[pste->side] ) {
            stats.cSafetyChange++;
            (pste-1)->reduced = false;
            pste->plyRem ++;
        }
#endif
#if 0
        if( (pste->kingsafety[pste->side]+25) < (pste-1)->kingsafety[pste->side] ) {
            stats.cSafetyChange++;
            preExtend = 1;
        }
#endif
    }
#else
    int nmPosVal = MaterialValue(pste) + valPAWN;
#endif

    // extCount is the net number of extension done (all extensions - all reductions done)
    int extCount = depth - pcon->ss.plyDepth + pste->plyRem; // extCount < 0 => reduced or wanted depth not reached yet
    int maxExt = depth - extCount;
    maxExt = maxExt * 3 / 4;

	//	Check hash table, in order to get "best" moves, and try to cut
	//	off.  Don't cut off within two plies of the root, otherwise the
	//	program might allow stupid rep draws.
	//
    PHASH hash_entry = 0;
#ifdef USE_HASH
    hash_entry = ProbeHashValue(pste);
    int hash_score = (hash_entry ? hash_entry->val : valMAX);
    if( hash_entry && (hash_entry->plyRem >= pste->plyRem) && (depth > 2) ) {
        PHASH phash = hash_entry;
        int val = phash->val;
#if 0
        if( val >= valMATEscore ) {
            val -= depth - 1;
        } else if( val <= -valMATEscore ) {
            val += depth - 1;
        }
#endif
		switch(phash->getBound())
        {
			case hashfEXACT:
                stats.Hcutoff ++;
                return val;
                break;
			case hashfBETA:
				{
                    if( val >= Beta ) {
                        stats.Hcutoff ++;
						return val;
                    }
// this can not work as is
                    if( val > Alpha ) {
                        stats.Halpha ++;
//                        Alpha = val;
                    }

					break;
				}
			case hashfALPHA:
				{
                    if( val <= Alpha ) {
                        stats.Hcutoff ++;
						return val;
                    }
                    // the next lines are sometimes good, sometimes bad...
                    if( val < Beta) {
                        stats.Hbeta++;
//                        Beta = val;
                    }
					break;
				}
            default:
                Assert(0);
        }
	}
#endif

    if( hash_entry )
        nm_threat = hash_entry->getThreat();

	//
	// Check alpha/beta window against a mate score
    // Thx for the idea
	//
#ifdef DISTANCE_PRUNING
#if 1
    // this is the lowest score I can get from here
    int i_am_mate_score = mateValue(depth + 1);
    if( i_am_mate_score > Alpha ) {
        Alpha = i_am_mate_score;
        // Beta is mate in n, best here is mate in n+m, so return
        if(i_am_mate_score >= Beta)
            return Beta; //i_am_mate_score;
//            return i_am_mate_score;
    }
#endif
#if 1
    // this is the best score we can get from here
    int i_will_mate_score =  -mateValue(depth);
    if( i_will_mate_score < Beta ) {
        Beta = i_will_mate_score;
        // if I set + 0 then I don't see the final move (
            // I will not see something better than Alpha so just return
        // Alpha is mate in n, best here is mate in n+m, so return
        if( i_will_mate_score <= Alpha )
            return Alpha;
//            return i_will_mate_score;
    }
#endif
#endif

    //
    // try recognizers 
    //
#ifdef USE_RECOGNIZER
    // the recognizers of bitbases where allready probed in the eval
    if( true 
        && (pste->bound != fail_bound) && pste->evaluated) {
        recog_result ret = {recog_bound(pste->bound), pste->val, recog_allways_valid};
        if( ret.bound == exact_bound )
            return ret.score;
        if( ret.bound == lower_bound ) {
            if( ret.score >= Beta )
                return ret.score;
//            if( ret.score > Alpha )
//                Alpha = ret.score;
        }
        if( ret.bound == upper_bound ) {
            if( ret.score <= Alpha )
                return ret.score;
            if( ret.score < Beta )
                Beta = ret.score;
        }
    }
#endif

    PHASHPawn phashp;
    EvalPawns(pcon, pste, phashp);

#ifdef NULL_MOVE
	//
	//	Do null-move search, which is a reduced depth that tries to
	//	figure out if the opponent has no threats.
	//
    // Note :
    // - if Beta is a mate score then it is not probable that I can do better without doing anything
    // - only done if not a PV (what about branches that will become a pv later ?)
    // - nm_threat caching does not work for an unknown reason (more nodes sometimes)
    // - material constraint reduced, I hope this works well with the verification search
    // - reductions are now allowed after an extension (so no test on extCount)
    // - if the stand pat score is < Beta then doing nothing will not raise it to Beta or more

    if ( 
        (!iaminCheck)
        && (pste->valPcUs > 0)
        && (pste->valPcThem > 0)
        && (nmPosVal > Beta 
//            || (pste->plyRem < 4)
            )
        && (Beta < valMATEscore)    // redondant with above test
        && (pste->plyRem >= 1)      // need more tests
        && (no_threat)
//        && (!phashp->pawnon6or7)
#ifdef HASH_THREAT
        && (!nm_threat)             // if we allready know that there is a threat...
#endif
        && (!pste->fNull)) {
        if( true ) {
            stats.NM1++;
            // don't forget to eval when the call to eval will be after bMove()
            int val = FThreat(pcon, pste, Alpha, Beta, nt);
            if( val >= Beta ) {
                // why not store result in hash ?
                // TODO check this: this does not work well on some position (more nodes)
                stats.NM2++;
                // verification search (for end game mostly)
                // still need to do NM at lower depth (higher remaining depth)
                const int Rv = 1 + 4;
                if(pste->plyRem >= Rv) {
                    stats.NM3++;
                    pste->fNull = true; // set or else we will loop :)
                    pste->plyRem -= Rv;
                    val = Search(pcon, pste, Alpha, Beta, nodeCUT /*nt*/);
                    pste->plyRem += Rv;
                }
            }
            if( val >= Beta ) {
#ifdef HASH_NM
                // store because we don't want to do this same search *again*
                RecordHashValue(pcon, pste, cmBest, val/*Beta*/, Alpha, Beta, depth, false);
#endif
                return val/*Beta*/;
            }
            if( true && (val < Alpha) 
//                && (pste->plyRem <= 7)
                && (pste-1)->reduced && 
                // there will be a re-search
                !( moveisVeryGoodCapture(&(pste+1)->argcmPv[0]) 
                    || (To((pste-2)->lastMove) == To((pste+1)->argcmPv[0].m))
                    || (To((pste-1)->lastMove) == From((pste+1)->argcmPv[0].m))
                    ) ) {
                    return valMIN;
            } else if( val <= -valMATEscore ) {
                // extend one ply if there is a mate threat
                // ok storing threat in hash does not work for some reason...
#ifdef HASH_THREAT
                nm_threat = true;
#endif
    #if 0
                sprintf(sz_score, "MATE threat in NM(%d)", val);
                VDumpLine(pcon, pste, sz_score);
                val = val;
    #endif
                // only setting king_attack so that we don't reduce the search
                // if we extend then we can search a lot deeper for no reason when
                // there is a mate but winning side does not play the winnin move
                king_attak = true;
//                preExtend = 1;
                // threat flag should be stored in hash too for next search
            }
        }
	}
#endif

#ifdef HASH_THREAT
    king_attak |= nm_threat;
#endif

	//	Mark this move so if I find it in a sub-tree I can detect a rep
	//	draw.
	//
	VSetRep(pcon, pste->hashkPc);
	//	Generate moves.
	//
    bGenMoves(pcon, pste);
	//
	//	I know how many reversible moves have been made up until now.
	//	I just tried to make another move.  If I am here, I didn't
	//	leave myself in check, so I succeeded.
	//
	//	If the number of reversible moves made before this is at
	//	least 50 for each side, this was a draw.
	//
    int lmc = legalMoveCount(pste);

	if ( (pste->plyFifty >= 50 * 2) && (lmc > 0) ) {
		pste->ccmPv = 0;
		VClearRep(pcon, pste->hashkPc);
		return 0;
	}
    // single move extension
    if( (lmc == 1) && iaminCheck)
        preExtend = 1;

#if 1
    Killers[depth+4].raz();
    MateKillers[depth+4].raz();
#endif

    PrepareSort( pcon, pste, depth, phashp );
	//	Mark the PV and hash moves so they sort first.
	//
	if (lastPV[depth].m != 0) {
		SortPV(pste, depth);
		lastPV[depth].m = 0;
	}
#ifdef USE_HASH
	if (hash_entry != NULL)
		SortHash(pste, hash_entry);
#endif
#ifdef USE_KILLERS
    SortKillers(pste, depth);
#endif

    int a = Alpha;
    int b = Beta;

    bool fFound = false;
    int  c_move = 0;

    bool f_prune = false;
    int  f_margin = valMAX;
    bool can_prune = false;

//#define FUTILITY_MARGIN (200)
//#define FUTILITY_MARGIN_PRE (250)
#define FUTILITY_MARGIN (150)
#define FUTILITY_MARGIN_PRE (250)
#define FUTILITY_MARGIN_RAZ (550)

#ifdef FUTILITY_PRUNING
    if( !iaminCheck 
//        && ((pste->valPcUs + pste->valPcThem) >= 2*valQUEEN)
        && ((pste->valPcUs + pste->valPcThem) >= 1*valQUEEN)
        && !phashp->pawnon6or7
        && (pste->plyRem <= 2) )
        can_prune = true;

    f_margin = FUTILITY_MARGIN * (pste->plyRem + 1);
//    if( pste->plyRem == 3 )
//        f_margin += 600;
#endif

    int depth2 = pste->plyRem * pste->plyRem;
    bool good_capt_done = false;
    pste->ttmove = hash_entry && (hash_entry->getMove() != 0);
	//
	//	Iterate through the move list, trying everything out.
	//
    foreachmove(pste,pcm) {
		//
		//	All moves are sorted on the cmk field
		//	cmk is set in PrepareSort() and in other sort functions

        SortNext(pste, pcm);
		//
		//	Make the move

		bMove(pcon, pste, pcm->m);
        if( moveisVeryGoodCapture(pcm) )
            good_capt_done = true;
        stats.c_int++;
	    posVal = ValEval(pcon, pste, a, b);

        int extension = preExtend;
        // Note : now I allow the reduction after the extension, 
        // this is needed for the check extension
        bool can_reduce = (extension) > 0 ? false : true;
        can_reduce = can_reduce && !king_attak;

        if( (pste + 1)->checkf ) {
            can_reduce = false;
            (pste+1)->checkdone[pste->side] = true;
#ifdef EXTEND_CHECK
            // extend everywhere in the tree
            // don't extend first check for side
            if( isFrontier(pste) || 
                (extCount < 77)
                && pste->checkdone[pste->side] && true
                ) { // 5,7,77
                // stop extending if I lead the score
                    if( a < valKNIGHT ) {
                        extension = 1;
                        stats.c_extcheck++;
                    }
            }
#else
            // only extend on check at the frontier
            if( isFrontier(pste) ) {
                // stop extending if I lead the score
                if( a < valKNIGHT )
                    extension = 1;
            }
#endif
        }
        if( moveisPawnMove(pcm) ) {
            // Pawn push to 7th
            if( moveto7thRank(pcm) ) {
//                extension = 1;
                can_reduce = false;
            } else if (((pste->valPcUs + pste->valPcThem) <= 1*valQUEEN))
                can_reduce = false;
        } else if( moveisPromotion(pcm) )
            can_reduce = false;

#ifdef SOME_PV_EXTENSION
        if( pste->ttmove && (pste-1)->ttmove ) {
            // node type can change to pv
            (pste+1)->c_ttpair++;
        }
#endif
        //  ==========================================================================
		//	Set up for recursion.
		//
		(pste + 1)->plyRem = pste->plyRem - 1;
        if( extension && (extCount < maxExt) ) {
            (pste + 1)->plyRem++;
        }
        //
		//	Recurse into normal search or quiescent search, as appropriate.
		//
        nodeType succNType;
        if(nt == nodeCUT )
            succNType = nodeALL;
        else if(nt == nodeALL || fFound)
            succNType = nodeCUT;

#ifdef PRUNE_SIBLING
#define MPD (50/1)  // 50/7
        if( can_reduce && fFound && isFrontier(pste)
            && !moveisGood(pcm)
            && !moveisCapture(pcm)
//            && ((!moveisCapture(pcm)) || (moveisBadCapture(pcm) && good_capt_done))
//            && !moveisPawnMove(pcm)
            ) {
            if( (posVal + MPD < a) ) {
                stats.FP_SIB++;
    		    bUndoMove(pcon, pste, pcm->m);
                int approx = posVal + (MPD >> 1);
                if( approx > bestVal ) {
                    bestVal = approx;
			        cmBest = *pcm;
			        pste->argcmPv[0] = cmBest;
			        pste->ccmPv = 1;
                }
                continue;
            }
        }
#endif

#ifdef FUTILITY_PRUNING

        if( fFound 
            && can_reduce && can_prune
            && !moveisPromotion(pcm) 
            && !moveisCapture(pcm)
#ifndef REDUCE_PAWN
            && !moveisPawnMove(pcm)
#endif
            ) 
            if( (posVal + f_margin) <= a ) {
                stats.FP_frontier++;
                int approx = posVal + (f_margin >> 2);
    		    bUndoMove(pcon, pste, pcm->m);
                if( approx > bestVal ) {
                    bestVal = approx;
			        cmBest = *pcm;
			        pste->argcmPv[0] = cmBest;
			        pste->ccmPv = 1;
                }
                continue;
            }
#endif

#ifdef USE_LMR
        // Note : LMR done on all non pv nodes moves
        //
        // move order is [H] [good capt] [equ capt] [killers] others [bad capt]
        if( 
            (c_move >= 1)
            && (!iaminCheck) && can_reduce
            && !moveisGood(pcm)
            && (pste->plyRem >= 1) 
//            && (pste->plyRem >= 2) 
            && (a > -valMATEscore)
            && !moveisPromotion(pcm) 
            // allow bad captures if good capture was seen
//            && !( moveisBadCapture(pcm) && !good_capt_done)
            && !moveisCapture(pcm)
#ifndef REDUCE_PAWN
            && !moveisPawnMove(pcm)
#endif
#ifndef USE_SIMPLE_LMR
            && History.can_reduce(pcm)
#endif
            ) {
            stats.LMRhint++;
            PHASH new_hash_entry = ProbeHashValue(pste+1);
            int new_hash_score = -(new_hash_entry ? new_hash_entry->val : valMAX);
            // it's not safe to reduce those moves
            // reduce if the score of the new position (from TT) is lower than alpha
            // search with a reduced depth
            stats.LMR++;
            pste->reduced = true;
            int reduc = 1;
#ifdef FUTILITY_PRUNING
            if( false && new_hash_entry && ((pste->prevCaptValue >= 300) ) && (pste->plyRem >= 2) && ((new_hash_score + (pste + 1)->plyRem * (valPAWN/2) + valBISHOP/2) <= a)  ) {
                stats.c_badscore++;
                reduc = 1 + (pste->plyRem / 2);
            }
#endif
		    (pste + 1)->plyRem -= reduc;
            val = -Search(pcon, pste + 1, - Beta, - a, succNType);
            // reduced was perhaps cleared
            if( val > a && pste->reduced) {
        		(pste + 1)->plyRem += reduc;
                pste->reduced = false;
                // re-search at normal depth
                stats.LMRreSearch++;
                val = -Search(pcon, pste + 1, - Beta, - a, succNType);
            }
            pste->reduced = false;
        }
        else
#endif
            val = -Search(pcon, pste + 1, - Beta, - a, succNType);

        bUndoMove(pcon, pste, pcm->m);
        if ((pcon->fAbort) || (pcon->fTimeout)) {
            VClearRep(pcon, pste->hashkPc);
            return 0;
        }
        //	We got a value back.  We unmade the move.  We're not dead.  Let's
        //	see how good this move was.  If it was >= "beta", it was so good
        //	that we don't need to search for anything better, so we'll leave.
        //
        //	If it was not >= "beta", but it was > "alpha", this is better than
        //	anything else we've found before, but not so good we have to
        //	leave.  These kinds of moves are actually quite rare.  If I find
        //	one of these, I have to store it in the PV (main-line) that I'm
        //	constructing.  This might end up being the main-line for the whole
        //	search, if it gets backed up all the way to the root.
        //
        if( val > bestVal ) {
            bestVal = val;
            if (val > a) {
//                if( nt == nodeALL)
//                    nt = nodeCUT;
                if (val >= Beta) {
                    //
                    //	This move failed high
                    //
                    //  Set the new killer move only if it was not marked as good allready
                    //  'Good' moves are allready sorted first so storing them in the killer
                    //  array is counter productive
                    // This was added to see mate search ; with the mate distance bound
                    // no pv was recorded 
                    cmBest = *pcm;
                    copyPV( pste, cmBest );
				    VClearRep(pcon, pste->hashkPc);
                    goto cutNode;
                }
                //	This move is between alpha and beta, which is actually pretty
                //	rare.  If this happens I have to add a PV move, and append the
                //	returned PV to it
                //
                cmBest = *pcm;
                copyPV( pste, cmBest );
                a = bestVal;
            }
        }
        fFound = true;
        c_move++;
        pste->ttmove = false;

#ifdef NULL_WINDOW
        b = a + 1;              // set the new null window
#endif
//        if( nt == nodeCUT && (c_move >= 7) && (bestVal < Alpha))
//            nt = nodeALL;
    }
    VClearRep(pcon, pste->hashkPc);

    if (!fFound) {
        // We have no move !!
        if (iaminCheck) {
            // Checkmate
            bestVal = mateValue(depth);
        } else {
            // Draw
            bestVal = 0;
        }
        // ==================================
        return bestVal;
        // ==================================
    }
    // the node type is not accurate, so I can not increment my (bad/good) move counters
    // in the move loop :(
    // if I have a move, then this is not an all nodes
cutNode:;

    if( bestVal > Alpha ) {
        pste->lastMove = cmBest.m;
        // for nm refutation
        (pste+1)->lastMove = cmBest.cmk;

        if( cmBest.cmk & cmkHASH )
            stats.c_cut_H++;
        else if( cmBest.cmk & cmkKILLER )
            stats.c_cut_killer++;
        else if( moveisGoodCapture(&cmBest) )
            stats.c_cut_capt++;
        else if( cmBest.cmk & cmkBADCAPT )
            stats.c_cut_badcapt++;
        else
            stats.c_cut_quies++;

        if( bestVal >= Beta ) {
            stats.cCUT++;
            if(iMove == 0)
                stats.cCUTrank1++;
            if(iMove < 5)
                stats.cCUTrankn++;
        } else
            stats.c_alpha_cut++;

#ifdef HISTORY
        if( bestVal >= Beta 
            && !moveisCapture(&cmBest) ) {   // don't update on captures
            // increment FH counter and classic history counter
            History.update( &cmBest, depth2, nt );
        }
#endif
#ifdef USE_KILLERS
        if( bestVal >= valMATEscore )
            MateKillers[depth].set(&cmBest);
        else if( !moveisGoodCapture(&cmBest) )
            Killers[depth].set(&cmBest);
#endif
    } else
        stats.c_all++;

    if( ! pcon->fTimeout )
        RecordHashValue(pcon, pste, cmBest, bestVal, LOWERBOUND, Beta, depth, nm_threat);
	return bestVal;
}



static int SearchPV(PCON pcon, PSTE pste, int Alpha, int Beta)
{
	int	val;
    pste->ccmPv = 0;
    pste->reduced = false;
	CM cmBest(0,0);
    int king_attacker = -1;
    int preExtend = 0;
    int bestVal = valMIN;
    int iaminCheck = pste->checkf;

    if (pste->plyRem < 0) {
        return SearchQ(pcon, pste,  Alpha, Beta);
    }

    int depth = int(pste - pcon->argste) + 1;
    if( depth >= (csteMAX-50) ) {
        return SearchQ(pcon, pste,  Alpha, Beta);
    }

	//	Increment global node counter, and figure out if it is time to call
	//	the interface callback, which is what allows for an interruptible
	//	search.  It's crucial that this be called now and then.
	//
	if (++pcon->ss.nodes >= pcon->ss.nodesNext) {
		//
		//	2000 nodes is maybe 1/50to 1/100 of a second.
		//
		pcon->ss.nodesNext = pcon->ss.nodes + 20000;
		VPrCallback(pcon);
		VCheckTime(pcon);
		if ((pcon->fAbort) || (pcon->fTimeout))
			return 0;
        if( pcon->smode == smodeTHINK || pcon->smode == smodeANALYZE ) 
            VPrSendUciCurrentLine(pcon, pste);
	}
	//
	//	I can return draws and cut off if I'm somewhere other than the
	//	first ply of depth.
	//
	//	Return draw score if this is a repeated node.
	//
	if (FRepSet(pcon, pste->hashkPc)) {
        stats.repCount++;
		return 0;
	}

    bool no_threat = true;
    // gathering stats
    int posVal = pste->val;
    if(!pste->evaluated )
	    posVal = ValEval(pcon, pste, Alpha, Beta);
//    int nmPosVal = posVal + valPAWN / 4;
//    int nmPosVal = posVal - 15;
    int nmPosVal = posVal;
    if( false && !pste->evalislazy ) {
        // if we have some hanging material
        if( pste->threat[pste->side] ) {
            no_threat = false;
            stats.c_threat++;
            if( (pste-1)->reduced ) {
                // there will be a re-search
                return valMIN;
            }
        }
    }
    if( !pste->evalislazy && !(pste-1)->evalislazy && (Alpha < valKNIGHT)) {
#if 0
        if( (pste-1)->reduced && (pste->kingsafety[pste->side]+35) < (pste-1)->kingsafety[pste->side] ) {
            stats.cSafetyChange++;
            (pste-1)->reduced = false;
            pste->plyRem ++;
        }
#endif
#if 0
        if( (pste->kingsafety[pste->side]+25) < (pste-1)->kingsafety[pste->side] ) {
            stats.cSafetyChange++;
            preExtend = 1;
        }
#endif
    }

    // extCount is the net number of extension done (all extensions - all reductions done)
    int extCount = depth - pcon->ss.plyDepth + pste->plyRem; // extCount < 0 => reduced or wanted depth not reached yet
    int maxExt = depth - extCount;
    maxExt = maxExt * 3 / 4;

	//	Check hash table, in order to get "best" moves, and try to cut
	//	off.  Don't cut off within two plies of the root, otherwise the
	//	program might allow stupid rep draws.
	//
    PHASH hash_entry = 0;
#ifdef USE_HASH
    hash_entry = ProbeHashValue(pste);
    int hash_score = (hash_entry ? hash_entry->val : valMAX);
#endif

	//
	// Check alpha/beta window against a mate score
    // Thx for the idea
	//
#ifdef DISTANCE_PRUNING
#if 1
    // this is the lowest score I can get from here
    int i_am_mate_score = mateValue(depth + 1);
    if( i_am_mate_score > Alpha ) {
        Alpha = i_am_mate_score;
        // Beta is mate in n, best here is mate in n+m, so return
        if(i_am_mate_score >= Beta)
            return Beta; //i_am_mate_score;
//            return i_am_mate_score;
    }
#endif
#if 1
    // this is the best score we can get from here
    int i_will_mate_score =  -mateValue(depth);
    if( i_will_mate_score < Beta ) {
        Beta = i_will_mate_score;
        // if I set + 0 then I don't see the final move (
            // I will not see something better than Alpha so just return
        // Alpha is mate in n, best here is mate in n+m, so return
        if( i_will_mate_score <= Alpha )
            return Alpha;
//            return i_will_mate_score;
    }
#endif
#endif

    //
    // try recognizers 
    //
#ifdef USE_RECOGNIZER
    // the recognizers of bitbases where allready probed in the eval
    if( true && (depth > 5)
        && (pste->bound != fail_bound) && pste->evaluated) {
        recog_result ret = {recog_bound(pste->bound), pste->val, recog_allways_valid};
        if( ret.bound == exact_bound )
            return ret.score;
        if( ret.bound == lower_bound ) {
            if( ret.score >= Beta )
                return ret.score;
        }
        if( ret.bound == upper_bound ) {
            if( ret.score <= Alpha )
                return ret.score;
            if( ret.score < Beta )
                Beta = ret.score;
        }
    }
#endif

    PHASHPawn phashp;
    EvalPawns(pcon, pste, phashp);

#ifdef USE_HASH
#ifdef USE_IID
    if( (hash_entry == NULL || (hash_entry->getMove() == 0)) && (pste->plyRem >= (4+0)) ) {
        // depth /2 gives worse results (more nodes)
//        int reduc = pste->plyRem / 2;
        int reduc = 2;
        pste->plyRem -= reduc;
        stats.c_iid++;
        int iidval = SearchPV(pcon, pste, Alpha, Beta);
#if 1
        /* frombeowulf : Re-search properly if the previous search failed low, so that we know we're getting
         * a good move, not just the move with the highest upper bound (which is essentially
         * random and depends on the search order.) */
        if( (iidval <= Alpha) ) {
//            iidval = Search(pcon, pste, /*valMIN*/iidval - valPAWN, Beta, nt);
            stats.c_iidbad++;
        }
#endif
        pste->plyRem += reduc;
        hash_entry = ProbeHashValue(pste);
        // Is it possible to not have a hash move here ? (how many times ?)
    }
#endif
#endif

	//	Mark this move so if I find it in a sub-tree I can detect a rep
	//	draw.
	//
	VSetRep(pcon, pste->hashkPc);
	//	Generate moves.
	//
    bGenMoves(pcon, pste);
	//
	//	I know how many reversible moves have been made up until now.
	//	I just tried to make another move.  If I am here, I didn't
	//	leave myself in check, so I succeeded.
	//
	//	If the number of reversible moves made before this is at
	//	least 50 for each side, this was a draw.
	//
    int lmc = legalMoveCount(pste);

	if ( (pste->plyFifty >= 50 * 2) && (lmc > 0) ) {
		pste->ccmPv = 0;
		VClearRep(pcon, pste->hashkPc);
		return 0;
	}
    // single move extension
    if( (lmc == 1) && iaminCheck)
        preExtend = 1;

#if 1
    Killers[depth+4].raz();
    MateKillers[depth+4].raz();
#endif

    PrepareSort( pcon, pste, depth, phashp );
	//	Mark the PV and hash moves so they sort first.
	//
	if (lastPV[depth].m != 0) {
		SortPV(pste, depth);
		lastPV[depth].m = 0;
	}
#ifdef USE_HASH
	if (hash_entry != NULL)
		SortHash(pste, hash_entry);
#endif
#ifdef USE_KILLERS
    SortKillers(pste, depth);
#endif

    int a = Alpha;
    int b = Beta;

#ifdef RECAPTURE_EXTENSION
    // margin to include Knight, Bishop and rook
    int prevCaptureValue    = pste->prevCaptValue;
    int prevCaptureValueMin = pste->prevCaptValue - 200;
    int prevCaptureValueMax = pste->prevCaptValue + 200;
    int prevCaptureSqr = pste->prevCaptSqr;
#endif

    bool fFound = false;
    int  c_move = 0;
    int depth2 = pste->plyRem * pste->plyRem;
    bool good_capt_done = false;
    pste->ttmove = hash_entry && (hash_entry->getMove() != 0);
	//
	//	Iterate through the move list, trying everything out.
	//
    foreachmove(pste,pcm) {
		//
		//	All moves are sorted on the cmk field
		//	cmk is set in PrepareSort() and in other sort functions

        SortNext(pste, pcm);
		//
		//	Make the move

		bMove(pcon, pste, pcm->m);
        if( moveisVeryGoodCapture(pcm) )
            good_capt_done = true;
        stats.c_int++;
	    posVal = ValEval(pcon, pste, a, b);

        int extension = preExtend;
        // Note : now I allow the reduction after the extension, 
        // this is needed for the check extension
        bool can_reduce = (extension) > 0 ? false : true;
        // only do some kind of extension on PV nodes
        bool ok_to_extend = true;

        if( (pste + 1)->checkf ) {
            can_reduce = false;
            (pste+1)->checkdone[pste->side] = true;
#ifdef EXTEND_CHECK
            // extend everywhere in the tree
            // don't extend first check for side
            if( isFrontier(pste) || 
                (extCount < 77)
                && pste->checkdone[pste->side] && true
                ) { // 5,7,77
                // stop extending if I lead the score
                    if( a < valKNIGHT ) {
                        extension = 1;
                        stats.c_extcheck++;
                    }
            }
#else
            // only extend on check at the frontier
            if( isFrontier(pste) ) {
                // stop extending if I lead the score
                if( a < valKNIGHT )
                    extension = 1;
            }
#endif
        }
        if( moveisPawnMove(pcm) ) {
            // Pawn push to 7th
            if( ok_to_extend && moveto7thRank(pcm) ) {
                extension = 1;
                can_reduce = false;
            } else if (((pste->valPcUs + pste->valPcThem) <= 1*valQUEEN))
                can_reduce = false;
        } else if( moveisPromotion(pcm) )
            can_reduce = false;

#ifdef SOME_PV_EXTENSION
        if( pste->ttmove && (pste-1)->ttmove ) {
            (pste+1)->c_ttpair++;
            if( (pste+1)->c_ttpair >= 2 
                && (hash_entry->getBound() != hashfALPHA) 
                ) {
                stats.c_ttmove++;
                (pste+1)->c_ttpair = 0;
                extension = 1;
                can_reduce = false;
            }
        }
#endif
#ifdef RECAPTURE_EXTENSION
        if( ok_to_extend && prevCaptureValue && Capt(pcm->m) && (To(pcm->m) == prevCaptureSqr) 
//            && moveisGoodCapture(pcm) 
            ) { 
            int matValue = recapMaterialValue(pcon, pcm);
            if( true || (matValue >= prevCaptureValueMin) && (matValue <= prevCaptureValueMax)) {
                stats.recapture++;
                extension = 1;
                can_reduce = false;
            }
        }
#endif
        //  ==========================================================================
		//	Set up for recursion.
		//
		(pste + 1)->plyRem = pste->plyRem - 1;
        if( extension && (extCount < maxExt) ) {
            (pste + 1)->plyRem++;
        }
        //
		//	Recurse into normal search or quiescent search, as appropriate.
		//
        if (!fFound) {
            val = -SearchPV(pcon, pste + 1, - Beta, - a);
            stats.c_pv++;
        } else {
            val = -Search(pcon, pste + 1, - b, - a, nodeCUT); // cut nodes
            if( val > a ) { // re-search
                stats.reSearch++;
                val = -SearchPV(pcon, pste + 1, - Beta, - a); // new PV
            }
        }

        bUndoMove(pcon, pste, pcm->m);
        if ((pcon->fAbort) || (pcon->fTimeout)) {
            VClearRep(pcon, pste->hashkPc);
            return 0;
        }
        //	We got a value back.  We unmade the move.  We're not dead.  Let's
        //	see how good this move was.  If it was >= "beta", it was so good
        //	that we don't need to search for anything better, so we'll leave.
        //
        //	If it was not >= "beta", but it was > "alpha", this is better than
        //	anything else we've found before, but not so good we have to
        //	leave.  These kinds of moves are actually quite rare.  If I find
        //	one of these, I have to store it in the PV (main-line) that I'm
        //	constructing.  This might end up being the main-line for the whole
        //	search, if it gets backed up all the way to the root.
        //
        if( val > bestVal ) {
            bestVal = val;
            if (val > a) {
                if (val >= Beta) {
                    //
                    //	This move failed high
                    //
                    //  Set the new killer move only if it was not marked as good allready
                    //  'Good' moves are allready sorted first so storing them in the killer
                    //  array is counter productive
                    // This was added to see mate search ; with the mate distance bound
                    // no pv was recorded 
                    cmBest = *pcm;
                    copyPV( pste, cmBest );
				    VClearRep(pcon, pste->hashkPc);
                    goto cutNode;
                }
                //	This move is between alpha and beta, which is actually pretty
                //	rare.  If this happens I have to add a PV move, and append the
                //	returned PV to it
                //
                cmBest = *pcm;
                copyPV( pste, cmBest );
                a = bestVal;
            }
        }
        fFound = true;
        c_move++;
        pste->ttmove = false;

#ifdef NULL_WINDOW
        b = a + 1;              // set the new null window
#endif
    }
    VClearRep(pcon, pste->hashkPc);

    if (!fFound) {
        // We have no move !!
        if (iaminCheck) {
            // Checkmate
            bestVal = mateValue(depth);
        } else {
            // Draw
            bestVal = 0;
        }
        // ==================================
        return bestVal;
        // ==================================
    }
    // the node type is not accurate, so I can not increment my (bad/good) move counters
    // in the move loop :(
    // if I have a move, then this is not an all nodes
cutNode:;

    if( bestVal > Alpha ) {
        pste->lastMove = cmBest.m;
        // for nm refutation
        (pste+1)->lastMove = cmBest.cmk;

        if( cmBest.cmk & cmkHASH )
            stats.c_cut_H++;
        else if( cmBest.cmk & cmkKILLER )
            stats.c_cut_killer++;
        else if( moveisGoodCapture(&cmBest) )
            stats.c_cut_capt++;
        else if( cmBest.cmk & cmkBADCAPT )
            stats.c_cut_badcapt++;
        else
            stats.c_cut_quies++;

        if( bestVal >= Beta ) {
            stats.cCUT++;
            if(iMove == 0)
                stats.cCUTrank1++;
            if(iMove < 5)
                stats.cCUTrankn++;
        } else
            stats.c_alpha_cut++;

#ifdef HISTORY
        if( bestVal >= Beta 
            && !moveisCapture(&cmBest) ) {   // don't update on captures
            // increment FH counter and classic history counter
            History.update( &cmBest, depth2, nodePV );
        }
#endif
#ifdef USE_KILLERS
        if( bestVal >= valMATEscore )
            MateKillers[depth].set(&cmBest);
        else if( !moveisGoodCapture(&cmBest) )
            Killers[depth].set(&cmBest);
#endif
    } else
        stats.c_all++;

    if( ! pcon->fTimeout )
        RecordHashValue(pcon, pste, cmBest, bestVal, LOWERBOUND, Beta, depth, false);
	return bestVal;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Full-width search.  This is the heart of the "chess playing" part of this
//	program.  It performs a recursive alpha-beta search.

static struct {
    char    move[12];
    int     c_nodes;
    int     hval;
} sortDebug[200];
static int c_sortDebug;

#define maxRootMoves    200
static struct _rootMove {
    char szmove[8];
    int score;
    int hint;
    U64 nodes;
    bool    evaluated;
    CM cm;
} rootMoves[maxRootMoves];
static int rootMovesCount;

static int sortScore( struct _rootMove & move) {
    return move.hint * 10000 + move.score;
}

int ValSearchRoot(PCON pcon, PSTE pste, int Alpha, int Beta)
{
    int a = Alpha;
    int b = Beta;
    int bestVal = valMIN;
    int	val;
//    int depth = 1;
    int depth = int(pste - pcon->argste + 1);
    pste->lastMove = 0;
    pste->reduced = false;
    pste->checkdone[B] = pste->checkdone[W] = false;
    pste->c_ttpair = 0;

    c_sortDebug = 0;

    //	Increment global node counter
    //
    ++pcon->ss.nodes;

    //	Stuff is doone a little differently at the root than at any other
    //	ply.
    //
    //
    //	Mark this move so if I find it in a sub-tree I can detect a rep
    //	draw.  I don't check to see if a rep is already set, because if
    //	we are here, we have to be making a move, not whining about a
    //	blown chance for a draw.  I also don't try to cut off from the
    //	hash table, for the same reason.  I want a move, not a score.
    //
    //	I do probe the table though, in case there is "best move" info
    //	in there.
    //
    PHASH hash_entry = ProbeHashValue(pste);

    VSetRep(pcon, pste->hashkPc);
    //
    //	If it's the first time we come here (depth == 0)
    //	generate the moves
    //
    if( pste->plyRem == 0 ) {
        // first time with this position
        bGenMoves(pcon, pste);
        pcon->ss.ccmLegalMoves = legalMoveCount(pste);
        rootMovesCount = pcon->ss.ccmLegalMoves;

        PHASHPawn phashp;
        EvalPawns(pcon, pste, phashp);
        PrepareSort( pcon, pste, depth, phashp );

        foreachmove(pste,pcm) {
            CmToSz( pcm, rootMoves[iMove].szmove );
            rootMoves[iMove].cm     = *pcm;
            rootMoves[iMove].hint   = 0;
            rootMoves[iMove].evaluated = false;
#if 0
            rootMoves[iMove].score  = - iMove; // keep the generation order
            if( Capt(pcm->m) )
                rootMoves[iMove].score -= 100;  // captures at the end
#else
            rootMoves[iMove].score = pcm->cmk & 0xffff;
            if( moveisCapture( pcm )) {
                if( moveisGoodCapture( pcm ) )
                    rootMoves[iMove].hint = 1;
                else
                    rootMoves[iMove].score = -1;
            }
#endif
            rootMoves[iMove].nodes  = 0;
        }
    }
    {
        // after the first time (first depth iteration)
        // sort the moves by node count
#ifdef USE_HASH
        PHASH phash = hash_entry;
        U32 hm = phash ? phash->getMove() : 0;
#endif
        for(int i = 0; i < rootMovesCount; i++) {
            PCM pcm = &rootMoves[i].cm;
            if ( rootMoves[i].cm.m == lastPV[1].m )
                rootMoves[i].hint = 10;
#ifdef USE_HASH
		    else if ( pcm->m == hm )
                rootMoves[i].hint = 1;
#endif
        }
    }
	//	Mark the PV and hash moves so they sort first.
	//
    // bubble sort
    for(int i = 0; i < rootMovesCount - 1; i++) {

        for(int j = i+1 ; j < rootMovesCount; j++)
            // most moves have the same score because of null window search
            if( (sortScore(rootMoves[j]) > sortScore(rootMoves[i]) ||
                (sortScore(rootMoves[j]) == sortScore(rootMoves[i])) && (rootMoves[j].nodes > rootMoves[i].nodes) )) {
                _rootMove swap = rootMoves[i];
                rootMoves[i] = rootMoves[j];
                rootMoves[j] = swap;

            }
    }

    CM cmBest(0,0);
	bool fFound = false;
	//
	//	Iterate through the move list, trying everything out.
	//
//	for (PCM pcm = pste->pcmFirst; pcm < (pste + 1)->pcmFirst; pcm++) {
    for(int i = 0; i < rootMovesCount; i++) {
        PCM pcm = &rootMoves[i].cm;
        pcm->cmk = i;
		//
		//
		//	Things are done differently at the root of the search.
		//
        SAN	san;
        //	Record information about which move I am searching ...
        //
        pcon->ss.icmSearching = i+1;
        CmToSan(pcon, pste, pcm, &san);
        SanToSz(&san, pcon->ss.aszSearching);
        CmToSz(pcm, pcon->ss.aszSearchSan);
        rootMoves[i].nodes = pcon->ss.nodes;
        VPrSendUciCurrMove( pcon->ss );

#ifdef _DEBUG
        {
            sortDebug[c_sortDebug].hval = 0;
            sortDebug[c_sortDebug].c_nodes = (int)pcon->ss.nodes;
            strcpy(sortDebug[c_sortDebug].move , pcon->ss.aszSearching);
            strcat(sortDebug[c_sortDebug].move , " ");
            if( pcm->cmk & cmkPV )
                strcat(sortDebug[c_sortDebug].move , "V");
            if( pcm->cmk & cmkHASH )
                strcat(sortDebug[c_sortDebug].move , "H");
            if( pcm->cmk & cmkKILLER )
                strcat(sortDebug[c_sortDebug].move , "K");
            if( pcm->cmk & cmkGOODCAPT )
                strcat(sortDebug[c_sortDebug].move , "G");
            if( pcm->cmk & cmkBADCAPT )
                strcat(sortDebug[c_sortDebug].move , "b");
            if( (pcm->cmk & cmkNONE) == cmkNONE)
                sortDebug[c_sortDebug].hval = pcm->cmk & ~0xfe000000;

        }
#endif
        //
        //	... then go make the move.
        //
        bMove(pcon, pste, pcm->m);
        bool can_reduce = true;
        int extension = 0;
        //
		//	Set up for recursion.
		//
		(pste + 1)->plyRem = pste->plyRem - 1;
        if( (pste + 1)->checkf )
            (pste+1)->checkdone[pste->side] = true;

#if 0
        if( (pste + 1)->checkf ) {
            can_reduce = false;
            // if I extend before max depth then I could get a doubled depth at the end
            // and it's a waste
            if( isFrontier(pste) ) {
                (pste + 1)->plyRem++;
                extension = 1;
            }
        }
#endif
		//
		//	Recurse into normal search or quiescent search, as appropriate.
		//
        nodeType succNType;
        if(fFound)
            succNType = nodeCUT;
        else 
            succNType = nodePV;

        if (!fFound) {
            val = -SearchPV(pcon, pste + 1, - Beta, - a);
        } else {
            val = -Search(pcon, pste + 1, - b, - a, succNType); // cut nodes
            if( val > a ) { // re-search
                stats.reSearch++;
                val = -SearchPV(pcon, pste + 1, - Beta, - a); // new PV
            }
        }

        // update the next sort key
        rootMoves[i].hint  = 0;
        if( ! rootMoves[i].evaluated )
            rootMoves[i].score = val;
        rootMoves[i].nodes = pcon->ss.nodes - rootMoves[i].nodes;

        // debugging stuff
        sortDebug[c_sortDebug].c_nodes = int(pcon->ss.nodes) - sortDebug[c_sortDebug].c_nodes;
        c_sortDebug++;

		bUndoMove(pcon, pste, pcm->m);
		if ((pcon->fAbort) || (pcon->fTimeout)) {
			VClearRep(pcon, pste->hashkPc);
			return 0;
		}
		//	We got a value back.  We unmade the move.  We're not dead.  Let's
		//	see how good this move was.  If it was >= "beta", it was so good
		//	that we don't need to search for anything better, so we'll leave.
		//
		//	If it was not >= "beta", but it was > "alpha", this is better than
		//	anything else we've found before, but not so good we have to
		//	leave.  These kinds of moves are actually quite rare.  If I find
		//	one of these, I have to store it in the PV (main-line) that I'm
		//	constructing.  This might end up being the main-line for the whole
		//	search, if it gets backed up all the way to the root.
        //
#if 0
        if( !strcmp(rootMoves[i].szmove, "a8a1" )) {
//                printf("# last bestval in pv is %d\n", pcon->ss.lastVal);
                printf("# Qa1 %d (%d) (%d %d)\n", val, bestVal, a, b);
                fflush(stdout);
//                copyPV( pste, *pcm );
//			    VDumpPv(pcon, pcon->ss.plyDepth, val, prsaNORMAL);
        }
#endif
        if( val > bestVal ) {
            bestVal = val;
		    if (val > a) {
			    if (val >= Beta) {
                    // Note : not reached if not doing aspiration search
				    //
				    //	This move failed high, so we are going to return beta,
				    //	but if we're sitting at the root of the search I will set
				    //	up a one-move PV (if this isn't already the move I'm
				    //	planning to make), so this move will be made if I run out
				    //	of time before resolving the fail-high.
				    //
                    copyPV( pste, *pcm );

				    //	This function needs the root moves generated, which
				    //	they are.
				    //
                    RecordHashValue(pcon, pste, *pcm, bestVal, LOWERBOUND, Beta, depth, false);
                    Beta = val;
				    VDumpPv(pcon, pcon->ss.plyDepth, Beta, prsaFAIL_HIGH);
				    VClearRep(pcon, pste->hashkPc);
				    return bestVal;
			    }
			    //	This move is between alpha and beta, which is actually pretty
			    //	rare.  If this happens I have to add a PV move, and append the
			    //	returned PV to it, and if I'm at the root I'll send the PV to
			    //	the interface so it can display it.
			    //
                rootMoves[i].score = val;
                rootMoves[i].evaluated = true;
			    cmBest = *pcm;
                copyPV( pste, cmBest );
			    a = bestVal;
			    //
			    //	This function needs the root moves generated, and they
			    //	are.
			    //	
			    VDumpPv(pcon, pcon->ss.plyDepth, a, prsaNORMAL);
#if 1 //_DEBUG
                stats.report();
#endif
		    }
        }
#ifdef NULL_WINDOW
        b = a + 1;              // null window
#endif
		fFound = true;
	}
#if 0
    printf("# SORT was ");
    for(int i = 0; i < c_sortDebug && i < 10; i++) {
        if( sortDebug[i].c_nodes < 10000 )
            printf("%s (%d) ", sortDebug[i].move, sortDebug[i].c_nodes );
        else
            printf("%s (%dk) ", sortDebug[i].move, sortDebug[i].c_nodes / 1000 );
        if( sortDebug[i].hval > 0 )
            printf("%d ", sortDebug[i].hval );
    }
    printf("\n");
#endif
	VClearRep(pcon, pste->hashkPc);
	if (!fFound) {
        // We have no move !!
		if (pste->checkf) {
            // Checkmate
			bestVal = mateValue(depth);
		} else
            // Draw
            bestVal = 0;
	}
    RecordHashValue(pcon, pste, cmBest, bestVal, LOWERBOUND, Beta, depth, false);
	return bestVal;
}
